/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:      CCM (obtain.C)
  
  Copyright: Damian J. J. Farnell

  Date:      Mon Feb 21 2000

  Updated:   Joerg Schulenburg 2004-2005

  Purpose:   Determine CCM LSUBm equations.

  ===================================================

  ---------------------------------------------------

  MODULE NAME: OBTAIN.C 

  PURPOSE:     `Obtains' the CCM equations for LSUBm

  ---------------------------------------------------

  Code checked on Thu Aug 3 11:08:54 BST 2000, DJJF.

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

extern void collect_term(int indices[6],double dummy,int id);
   void collect_gse_term(int indices[6],double dummy,int id);
extern int subsetClusters(int (*config)[5], int length); // 1 if ok, 0 if not

int radius=0; // only used within obtain.C, better (faster) locally?
int h_id=-1;  // file id for matrix portion, 0=gse, 1=terms, 2=expects
int minX, maxX, minY, maxY, minZ, maxZ; // range for additional flips

// defined in collect.C
extern unsigned long
 nterms_in1,  nterms_in2,
 nterms_out1, nterms_out2;

// see at bottom for obtain(void)

// Jun08 Joerg new: find DescriptorIndex from Cluster (one call)
// replacing three lines by one (ToDo: call subset too?)
//   return <= 0 (not found), >=1 (found) 
int findClusterIndex(int order, int (*config0)[5]) {
  static int descriptor[MAX_SPIN_LIST];
  if (order > Params.n)                return 0; // faster? 
  if (!subsetClusters(config0, order)) return 0;   
  findDescriptorOfCluster(order, config0, descriptor); // bookKeeping.C
  return findDescriptorIndex(descriptor, spinConfigurations,
                             NumberConfigurations);
}

//  split configuration conf0 into conf1 and conf2 according to split-array
//  take sites from conf0[1..len0] and append (!) to conf1[len1..] + conf2
// static // extern for obtainGeneralisedExpectation()
//   use:
//      for (int j=1; j<=len; j++) split[j]=0;
//      for (ovfl=0;!ovfl;) {
//         ocfl = split_binary(conf0,len0,split,...);
//         // split to 000 001 010 011 100 101 110 111 nosplit+break
//         ...
//      }
int  split_binary(int (*conf0)[5], int  len0, int *split0,
                  int (*conf1)[5], int *len1,
                  int (*conf2)[5], int *len2){
  int i, *pd, *ps;
  for (i=1; i<=len0; i++) {
    ps=conf0[i];
    if (split0[i]) { (*len2)++; pd=conf2[*len2]; } // destination config 1
    else           { (*len1)++; pd=conf1[*len1]; } // destination config 0
    pd[1]=ps[1]; // x
    pd[2]=ps[2]; // y
    pd[3]=ps[3]; // z
    pd[4]=ps[4]; // uc
  }
  // count up (dual)
  for(i=1;i<=len0;i++) { split0[i]^=1; if (split0[i]) break; }
  return ((i>len0)?1:0); // counter overflow
}

int  split_trimary(int (*conf0)[5], int  len0, int *split0,
                   int (*conf1)[5], int *len1,
                   int (*conf2)[5], int *len2,
                   int (*conf3)[5], int *len3){
  int i, *pd, *ps;
  for (i=1; i<=len0; i++) {
    ps=conf0[i];
    switch (split0[i]) {
      case 2:  {(*len3)++; pd=conf3[*len3];};break; // destination config 2
      case 1:  {(*len2)++; pd=conf2[*len2];};break; // destination config 1
      default: {(*len1)++; pd=conf1[*len1];};break; // destination config 0
    }
    pd[1]=ps[1]; // x
    pd[2]=ps[2]; // y
    pd[3]=ps[3]; // z
    pd[4]=ps[4]; // uc
  }
  // count up (trimal)
  for(i=1;i<=len0;i++) { split0[i]=(split0[i]+1)%3; if (split0[i]) break; }
  return ((i>len0)?1:0); // counter overflow
}

// Joerg ToDo: bonds or expect is a bad name, better use h_component, Jun08

/* test if site s1 and site s2 are neighbours k -> m */
// static
int test_k_to_m_neighbour(struct hamiltonian_defn hh[],
                          int term, int s1[5], int s2[5]) {
  int are_they_neighbours=0;
  // struct hamiltonian_defn hh[]=&Hamiltonian[bonds][map62[term]];
  int index=s1[4];
  int j;
  // neighbourhood of s2 to s1 (can be directional!)
  for (j=1;j<=hh[map62[term]].k_to_m_number[index];j++) {
	/* k ---> m                      uc   nn xyzc */
	if (s2[1] == s1[1] + hh[map62[term]].k_to_m[index][j][0] /* x */
	 && s2[2] == s1[2] + hh[map62[term]].k_to_m[index][j][1] /* y */
	 && s2[3] == s1[3] + hh[map62[term]].k_to_m[index][j][2] /* z */
	 && s2[4] ==         hh[map62[term]].k_to_m[index][j][3] /* uc */ )
    { are_they_neighbours=j; break; }
  }
  return are_they_neighbours;
}
/* test if site s1 and site s2 are neighbours m -> k */
// static
int test_m_to_k_neighbour(struct hamiltonian_defn hh[],
                          int term, int s1[5], int s2[5]) {
  int are_they_neighbours=0;
  // struct hamiltonian_defn hh[]=&Hamiltonian[bonds][map62[term]];
  int index=s1[4];
  int j;
  // neighbourhood of s2 to s1 (can be directional!)
  for (j=1;j<=hh[map62[term]].m_to_k_number[index];j++) {
	/* m ---> k                      uc   nn xyzc */
	if (s2[1] == s1[1] + hh[map62[term]].m_to_k[index][j][0] /* x */
	 && s2[2] == s1[2] + hh[map62[term]].m_to_k[index][j][1] /* y */
	 && s2[3] == s1[3] + hh[map62[term]].m_to_k[index][j][2] /* z */
	 && s2[4] ==         hh[map62[term]].m_to_k[index][j][3] /* uc */ )
    { are_they_neighbours=j; break; }
  }
  return are_they_neighbours;
}

// better extra byte for num flips of this site?
int to_much_flips(int (*conf0)[5], int len0, int site0[5])
{
  // check that we dont have to much flips on site site0
  int m, numberFlips=0;
  for(m=1;m<=len0;m++)
    if (  site0[1]==conf0[m][1]
       && site0[2]==conf0[m][2]
       && site0[3]==conf0[m][3]
       && site0[4]==conf0[m][4] ) numberFlips++;
  if (numberFlips > UnitCell.s2[ site0[4] ]) return 1;
  return 0;
}

// --------------------------------- terms --------------------------
// s^z s^z : all s 
// == term3 backwards
double term2(int config_index, int length, int (*config0)[5],
             struct hamiltonian_defn hh[], int term)
{
  /*      
     TERM 01.02: F(m) . S(+)(m) . S(z)(k)  Checked 3-7-2000.
     TERM 01.03: F(k) . S(+)(k) . S(z)(m)
  */
  
//  int i,j,
  int k;
  double multiplier=0.0;
  int neighbour;
  int index;
  int new_index;
  double sum;
  sum=0.0000;
  
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  // get the multiplying coefficient correct first!

  for(k=1;k<=length;k++) {
    
    index = config0[k][4]; // for every site ...
    
    for(neighbour=1;
        neighbour<=((term==2) ? hh[map62[term]].m_to_k_number[index]
                    /* 3 */   : hh[map62[term]].k_to_m_number[index]);
	neighbour++) {
      
      new_index =((term==2) ? hh[map62[term]].m_to_k[index][neighbour][3]
                  /* 3 */   : hh[map62[term]].k_to_m[index][neighbour][3]);
      
      multiplier += UnitCell.s2[new_index]
                   *hh[map62[term]].coefficient[index][neighbour];
    }
  }
  
  // -.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
  
  {
    int indices[7]={0};
    double dummy;
    indices[0]=config_index;
    indices[1]=term;
    indices[2]=config_index;
    indices[3]=0;
    indices[4]=0;
    indices[5]=0;
    sum=dummy=-0.5*multiplier;
    collect_term(indices,dummy,h_id); // JS0902: called once per config now
  }
  return sum;
}

// term3 = term2 backwards

double term4(int config_index, int length, int (*config0)[5],
             struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 4: H(k,m) . S(+)(k) . S(+)(m).
  */
  int k1, k2;
  int are_they_neighbours, index;
  double sum=0.0;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  for(k1=1;k1<=length;k1++)
  for(k2=1;k2<=length;k2++)
  if (k1 != k2) {               // loop over nearest neighbours

    index = config0[k1][4];
    are_they_neighbours= 
      test_k_to_m_neighbour(hh, term, config0[k1], config0[k2]);
    if (!are_they_neighbours) continue;
    
    sum += hh[map62[term]].coefficient[index][are_they_neighbours];
  }
  {
    int indices[7]={0};
    double dummy;
    indices[0]=config_index;
    indices[1]=term;
    indices[2]=config_index;
    indices[3]=0;
    indices[4]=0;
    indices[5]=0;
    dummy=sum;
    collect_term(indices,dummy,h_id); // JS0902: called once per cfg now
  }
  return sum;
}

double term5(int config_index, int length, int (*config0)[5],
             struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 5: F(k) . F(m) . S(+)(k) . S(+)(m).
  */

  int split0[MAX_LSUBM+1], i1;
  int counter;
  int (*conf0)[5];
  int (*conf1)[5];
  int (*conf2)[5];
  int k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  // double multiplier;
  int conf1_len;
  int conf2_len;
  double sum;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  sum=0.0000;

  conf0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k1=1;k1<=length;k1++)
  for(k2=1;k2<=length;k2++)
  if (k2 != k1) {
    
    for(m=1;m<5;m++) conf1[1][m]=config0[k1][m]; 
    for(m=1;m<5;m++) conf2[1][m]=config0[k2][m];
    
    // loop over nearest neighbours
    
    index = conf1[1][4];
    are_they_neighbours=
      test_k_to_m_neighbour(hh,term, conf1[1], conf2[1]);
    
    // conf1[1] and conf2[1] need to be nearest-neighbours 
    
    if (!are_they_neighbours) continue; 
    
    for(counter=0,l=1;l<=length;l++) {
      if (l != k1 && l != k2 ) {
        counter++;
        conf0[counter][1]=config0[l][1]; // x
        conf0[counter][2]=config0[l][2]; // y
        conf0[counter][3]=config0[l][3]; // z
        conf0[counter][4]=config0[l][4]; // uc
      }
    }
    
    for (i1=1;i1<=counter;i1++) split0[i1]=0; // initiate splitter
    for (i1=0;!i1;) {

      conf1_len=1;
      conf2_len=1;
      
      // now we have next binary code 000 -> 001 -> 010 ...
      i1=split_binary(conf0,  counter, split0,
                      conf1,  &conf1_len,  // P(k)
                      conf2,  &conf2_len); // F(m)
    
      descriptorIndex1=findClusterIndex(conf1_len, conf1);
      if (descriptorIndex1 <= 0) continue;

      descriptorIndex2=findClusterIndex(conf2_len, conf2);
      if (descriptorIndex2 <= 0) continue;
      
      // rhs this time!
      
      // if descriptorIndex is 0, it means no matching cluster
      // from the `fundamental' set could be found and so 
      // this case is ignored -- i.e., not saved to disk.
      
      {
        int indices[7]={0};
        double dummy;
        indices[0]=config_index;
        indices[1]=term;
        indices[2]=descriptorIndex1;
        indices[3]=descriptorIndex2;
        indices[4]=0;
        indices[5]=0;
        dummy=hh[map62[term]].coefficient[index][are_they_neighbours];
        collect_term(indices,dummy,h_id);
      }
      
    }
  }
  free(conf0);
  free(conf1);
  free(conf2);
  return sum;
}

// term14 = term6 backwards
double term6(int config_index, int length, int (*string)[5],
             struct hamiltonian_defn hh[], int term) {
  /*
   TERM 6:  02.06  =  S(+)(m) . S(z)(k)  =  z+ [0]
   TERM 14: 04:14  =  S(+)(k) . S(z)(m)  =  +z [0]
  */
  
  int neighbour;
  double multiplier;
  int index;
  double sum;
  int new_index;
  
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  sum=0.0;
  
  if (length != 1)
    return 0.0;
  
  index=string[1][4];
        
  for(neighbour=1;
      neighbour<=((term==6)?hh[map62[term]].m_to_k_number[index]:
                   /* 14 */ hh[map62[term]].k_to_m_number[index]);
      neighbour++) {
    
    // nearest neighbour is at x,y,z, cell put information
    // into configuration and exit.
    
    if (term==6) {
      new_index=hh[map62[term]].m_to_k[index][neighbour][3];
    } else { /* term==14 */
      new_index=hh[map62[term]].k_to_m[index][neighbour][3];
    }
    
    // cout << "new Index = " << new_index << endl;
    
    multiplier=-0.5*UnitCell.s2[new_index];
    
    sum+=multiplier*hh[map62[term]].coefficient[index][neighbour];
  }
  {
    int indices[7]={0};
    double dummy;
    indices[0]=config_index;
    indices[1]=term;
    indices[2]=0; // !Special Case! No CCM ket-state coeffs here!
    indices[3]=0; // This is checked for in solve.C
    indices[4]=0;
    indices[5]=0;
    dummy=sum;
    collect_term(indices,dummy,h_id);
  }
  return sum;
}

// s^z s^ + all s
// term15 = term7 backwards
double term7(int config_index, int length, int (*config0)[5],
             struct hamiltonian_defn hh[], int term) {
  /*
     TERM 7:  F(k) . S(+)(k) . S(+)(m).
     TERM 15: F(m) . S(+)(m) . S(+)(k).
  */
  
  int i, j, l;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex;
  double multiplier;
  int (*conf)[5];
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  conf=  (int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  // configuration 1
  // (length-2)!     -- permutation factor because we sum over k1 and k2.
  // (length-1)      -- factor from F(). Note that s_k2^+ `captures' one
  //                    of the spin flips but does not contribute to F().
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k1 != k2) {  // ToDo: better k1 < k2 and factor 2 ??? Joerg
	
	for(l=1;l<5;l++)    conf[1][l]=config0[k1][l];
	for(l=1;l<5;l++) tempConfig[l]=config0[k2][l];
	
	// loop over nearest neighbours
	
        index = conf[1][4];

        are_they_neighbours=
          ((term==7)?test_k_to_m_neighbour(hh,term, conf[1], tempConfig):
	  /* 15 */   test_m_to_k_neighbour(hh,term, conf[1], tempConfig));

	// conf[1][] and tempConfig[] need to be nearest-neighbours 
	
	// now load the rest into conf
	
	for(counter=1,i=1;i<=length;i++)
	if (i != k2 && i != k1) {
	  counter++;
	  for(l=1;l<5;l++) conf[counter][l]=config0[i][l];
	}
	
	if (counter != (length-1)) exit(1);
	
	if (!are_they_neighbours) continue;
        
        descriptorIndex=findClusterIndex(length-1,conf);
        
        if (descriptorIndex > 0) {
          int indices[7]={0};
          double dummy;
          indices[0]=config_index;
          indices[1]=term;
          indices[2]=descriptorIndex;
          indices[3]=0;
          indices[4]=0;
          indices[5]=0;
          dummy=multiplier*hh[map62[term]].coefficient[index][are_they_neighbours];
          collect_term(indices,dummy,h_id);
        }
      }
    }
  }
  free(conf);
  return sum;
}

// term8 = term17 backwards
double term8(int config_index, int length, int (*config0)[5],
             struct hamiltonian_defn hh[], int term) {
  /*
     TERM  8: -2 * F(m) . S(z)(k) . S(z)(m)
     TERM 17: -2 * F(m) * S(z)(k) * S(z)(m)     
     Note that term8 generally cancels with term 17.......
  */

  int j;
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  int x,y,z,cell;
  int index;
  
//  int numberFlips, maxNumberFlips;
  int nn[5];

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */
  // Oct05: configs with length+1 does not generate terms, so we skip it
  if (length+1>Params.n) return 0;

  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
	  // order=length+1;	            // this is the `m'    
	  
	  config0[length+1][1]=x;
	  config0[length+1][2]=y;
	  config0[length+1][3]=z;
	  config0[length+1][4]=cell;
	  
	  index=cell;

	  // check that we don't have too many spins at site `m'.
          // ToDo: replace by (x,y,z,uc,numflips)-tupel, JS0903
          if (to_much_flips(config0, length+1, config0[length+1]))
              continue;
          
          for(j=1;j<=((term==8)?hh[map62[term]].m_to_k_number[index]:
                     /* 17 */   hh[map62[term]].k_to_m_number[index]);j++) {
            
            // find the s_k^z, and so use
            
            if (term==8) {
              nn[1]=config0[length+1][1]+hh[map62[term]].m_to_k[index][j][0];
              nn[2]=config0[length+1][2]+hh[map62[term]].m_to_k[index][j][1];
              nn[3]=config0[length+1][3]+hh[map62[term]].m_to_k[index][j][2];
              nn[4]=                     hh[map62[term]].m_to_k[index][j][3];
            } else { /* term==17 */
              nn[1]=config0[length+1][1]+hh[map62[term]].k_to_m[index][j][0];
              nn[2]=config0[length+1][2]+hh[map62[term]].k_to_m[index][j][1];
              nn[3]=config0[length+1][3]+hh[map62[term]].k_to_m[index][j][2];
              nn[4]=                     hh[map62[term]].k_to_m[index][j][3];
            }
            
            temp_multiplier=-2*0.5*UnitCell.s2[config0[length+1][4]]*
                               0.5*UnitCell.s2[nn[4]];
            
            descriptorIndex=findClusterIndex(length+1,config0);
            if (descriptorIndex <= 0) continue;
            
            {
                int indices[7]={0};
                double dummy;
                indices[0]=config_index;
                indices[1]=term;
                indices[2]=descriptorIndex;
                indices[3]=0;
                indices[4]=0;
                indices[5]=0;
                dummy=temp_multiplier*hh[map62[term]].coefficient[index][j];
                collect_term(indices,dummy,h_id);
            }
            
          }
	}
      }
    }    
  }
  return sum;  
}

// term16 = term9 backwards
double term9(int config_index, int length, int (*string)[5],
             struct hamiltonian_defn hh[], int term) {
  /*
     TERM 9: -2 * H(k)(m) . S(+)(k) . S(z)(m) // remove (l) = length
    TERM 16: -2 * H(m)(k) . S(+)(m) . S(z)(k)     
  */

  int i,j,k;
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int (*config0)[5];
  int index;
  int nn[5];
  int neighbour;
//  int maxNumberFlips;
//  int numberFlips;
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  config0=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      config0[i][j]=string[i][j];

  multiplier=1.0;
  //  for(i=1;i<length;i++) multiplier*=i; // permutation factor
  // multiplier*=(length+1)*length; // factor from H()
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=config0[k][4];
    
    for(neighbour=1;
        neighbour<=((term==9)?hh[map62[term]].k_to_m_number[index]:
                     /* 16 */ hh[map62[term]].m_to_k_number[index]);
	neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
	
      if (term==9) {
        /* k ---> m */
        nn[1]=config0[k][1]+hh[map62[term]].k_to_m[index][neighbour][0];
        nn[2]=config0[k][2]+hh[map62[term]].k_to_m[index][neighbour][1];
        nn[3]=config0[k][3]+hh[map62[term]].k_to_m[index][neighbour][2];
        nn[4]=                    hh[map62[term]].k_to_m[index][neighbour][3];
      } else { /* term==16 */
        /* m ---> k */
        nn[1]=config0[k][1]+hh[map62[term]].m_to_k[index][neighbour][0];
        nn[2]=config0[k][2]+hh[map62[term]].m_to_k[index][neighbour][1];
        nn[3]=config0[k][3]+hh[map62[term]].m_to_k[index][neighbour][2];
        nn[4]=              hh[map62[term]].m_to_k[index][neighbour][3];
      }

      config0[length+1][1]=nn[1];
      config0[length+1][2]=nn[2];
      config0[length+1][3]=nn[3];
      config0[length+1][4]=nn[4];

      if (to_much_flips(config0, length+1, config0[length+1]))
         continue;

      descriptorIndex=findClusterIndex((length+1),config0);
        
      // UnitCell.s2[] from coefficient factors in the term 8

      temp_multiplier=multiplier*UnitCell.s2[nn[4]];
        
      if (descriptorIndex > 0) {
        int indices[7]={0};
        double dummy;
        indices[0]=config_index;
        indices[1]=term;
        indices[2]=descriptorIndex;
        indices[3]=0;
        indices[4]=0;
        indices[5]=0;
        dummy=temp_multiplier*hh[map62[term]].coefficient[index][neighbour];
        collect_term(indices,dummy,h_id);
      }
    }
  }
  free(config0);
  return sum;
}  

// term20 = term10 backwards
double term10(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term) {

 /*
   TERM 10: - F(m) . F(m) . S(+)(m) . S(z)(k)
   TERM 20: - F(k) . F(k) . S(+)(k) . S(z)(m)
  */

  int split0[MAX_LSUBM+1], i1;
  int k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int index;
  int nn[5];
  int neighbour;
  int (*conf0)[5], (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  conf0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) conf1[1][m]=config0[k][m]; // F(m)
    for(m=1;m<5;m++) conf2[1][m]=config0[k][m]; // F(m)

    for(counter=0,l=1;l<=length;l++)
    if (l != k) {
        counter++;
        conf0[counter][1]=config0[l][1]; // x
        conf0[counter][2]=config0[l][2]; // y
        conf0[counter][3]=config0[l][3]; // z
        conf0[counter][4]=config0[l][4]; // uc
    }
    
    for (i1=1;i1<=counter;i1++) split0[i1]=0; // initiate splitter
    for (i1=0;!i1;) {

      conf1_len=1;
      conf2_len=1;
      
      // now we have next binary code 000 -> 001 -> 010 ...
      i1=split_binary(conf0,  counter, split0,
                      conf1,  &conf1_len,  // F(m)
                      conf2,  &conf2_len); // F(m)

      index=conf1[1][4];
      
      for(neighbour=1;
          neighbour<=((term==10)?hh[map62[term]].m_to_k_number[index]:
                                 hh[map62[term]].k_to_m_number[index]);
          neighbour++) {
	
	// find s_k^z and so use
	if (term==10) {
  	  /* m ---> k */
	  nn[1]=conf1[1][1]+hh[map62[term]].m_to_k[index][neighbour][0];
	  nn[2]=conf1[1][2]+hh[map62[term]].m_to_k[index][neighbour][1];
	  nn[3]=conf1[1][3]+hh[map62[term]].m_to_k[index][neighbour][2];
	  nn[4]=            hh[map62[term]].m_to_k[index][neighbour][3];
	} else { /* term==20 */
  	  /* k ---> m */
	  nn[1]=conf1[1][1]+hh[map62[term]].k_to_m[index][neighbour][0];
	  nn[2]=conf1[1][2]+hh[map62[term]].k_to_m[index][neighbour][1];
	  nn[3]=conf1[1][3]+hh[map62[term]].k_to_m[index][neighbour][2];
	  nn[4]=            hh[map62[term]].k_to_m[index][neighbour][3];
        }

        // JS ToDo: move this expensive findClustrIndex outside this loop!
	descriptorIndex1=findClusterIndex(conf1_len,conf1);
	if (descriptorIndex1 <= 0) continue;
	
	descriptorIndex2=findClusterIndex(conf2_len,conf2);
	if (descriptorIndex2 <= 0) continue;
	
	multiplier=0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^- \tilde s^z
	
        {
          int indices[7]={0};
          double dummy;
          indices[0]=config_index;
          indices[1]=term;
          indices[2]=descriptorIndex1;
          indices[3]=descriptorIndex2;
          indices[4]=0;
          indices[5]=0;
          dummy=multiplier*hh[map62[term]].coefficient[index][neighbour];
          collect_term(indices,dummy,h_id);
        }
      }
    }
  }
  free(conf0);
  free(conf1);
  free(conf2);
  return sum;
}

double term11(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term) {
  /*
     TERM 11: -2 * H(k)(m) . F(m) . S(+)(k) . S(+)(m)
     TERM 19: -2 * H(k)(m) . F(k) . S(+)(k) . S(+)(m)
   */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int j, k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=config0[k1][m]; // H(k,m) <-- k here
	for(m=1;m<5;m++) conf2[1][m]=config0[k2][m]; // F(m)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          ((term==11)?test_k_to_m_neighbour(hh,term, conf1[1], conf2[1]):
	   /* 19 */   test_m_to_k_neighbour(hh,term, conf1[1], conf2[1]));

	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++) conf1[2][m]=config0[k2][m];
	  
	  for(counter=0,l=1;l<=length;l++)
	  if (l != k1 && l != k2 ) {
            counter++;
	    tempArray[counter]=l;
	  }
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=2;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {  // H(k,m)
		conf1_len++;  temp=tempArray[j];
		for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
	      } else {        // F(m)
		conf2_len++;  temp=tempArray[j];
		for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findClusterIndex(conf1_len,conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findClusterIndex(conf2_len,conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    multiplier=-2.0;
	    
	    // coeff from \tilde s^z \tilde s^-
	    
            {
              int indices[7]={0};
              double dummy;
              indices[0]=config_index;
              indices[1]=term;
              indices[2]=descriptorIndex1;
              indices[3]=descriptorIndex2;
              indices[4]=0;
              indices[5]=0;
              dummy=multiplier*hh[map62[term]].coefficient[index][are_they_neighbours];
              collect_term(indices,dummy,h_id);
            }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

// term12 = term18 backwards
double term12(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term) {
  /*
     TERM 12: -2 * F(k) . F(m) . S(+)(k) . S(z)(m)
     TERM 18: -2 * F(m) . F(k) . S(+)(m) . S(z)(k)
  */

  int j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int index;
  int nn[5];
  int neighbour;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
//  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) conf1[1][m]=config0[k][m]; // F(k)

    for(counter=0,m=1;m<=length;m++)
    if (m != k) {
	counter++;
	tempArray[counter]=m;
    }
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
        temp=tempArray[j];
	if (spin[j]) {  // F(k)
	  conf1_len++;
	  for (m=1;m<5;m++)  conf1[conf1_len][m]=config0[temp][m];
	} else {        // F(m)
	  conf2_len++;
	  for (m=1;m<5;m++)  conf2[conf2_len][m]=config0[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;
          neighbour<=((term==12)?hh[map62[term]].k_to_m_number[index]:
                      /* 18 */   hh[map62[term]].m_to_k_number[index]);
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into config0 and exit.
	if (term==12) {
	  /* k ---> m */
	  nn[1]=conf1[1][1]+hh[map62[term]].k_to_m[index][neighbour][0];
	  nn[2]=conf1[1][2]+hh[map62[term]].k_to_m[index][neighbour][1];
	  nn[3]=conf1[1][3]+hh[map62[term]].k_to_m[index][neighbour][2];
  	  nn[4]=            hh[map62[term]].k_to_m[index][neighbour][3];
        } else { /* term==18 */
	  /* m ---> k */
	  nn[1]=conf1[1][1]+hh[map62[term]].m_to_k[index][neighbour][0];
	  nn[2]=conf1[1][2]+hh[map62[term]].m_to_k[index][neighbour][1];
	  nn[3]=conf1[1][3]+hh[map62[term]].m_to_k[index][neighbour][2];
  	  nn[4]=            hh[map62[term]].m_to_k[index][neighbour][3];
        }
	
        conf2[conf2_len+1][1]=nn[1];
        conf2[conf2_len+1][2]=nn[2];
        conf2[conf2_len+1][3]=nn[3];
        conf2[conf2_len+1][4]=nn[4];
        
        if (to_much_flips(conf2,conf2_len+1,conf2[conf2_len+1]))
          continue;
        
        descriptorIndex1=findClusterIndex(conf1_len,conf1);
        if (descriptorIndex1 <= 0) continue;
        
        descriptorIndex2=findClusterIndex(conf2_len+1,conf2);
        if (descriptorIndex2 <= 0) continue;
        
        multiplier=1.0;
        // for(i=1;i<=conf1_len-1;i++) multiplier*=i; 
        // multiplier*=conf1_len;
        
        // F(k)
        // (length-1)!       -- because we sum over k
        // length            -- internal factor in F(k)
        
        // for(i=1;i<=conf2_len+1-1;i++) multiplier*=i;
        // multiplier*=(conf2_len+1); 
        
        // F(m)
        // (length-1)!       -- permutation factor from F(m)
        // length            -- internal factor in F(m) because `m' 
        //                      is `virtual'. (Note that summation over 
        //                      n.n. gives m thus length ->length+1.)
        
        multiplier=2*multiplier*0.5*UnitCell.s2[nn[4]];
        
        // coeff from \tilde s^- \tilde s^z
        
        {
          int indices[7]={0};
          double dummy;
          indices[0]=config_index;
          indices[1]=term;
          indices[2]=descriptorIndex1;
          indices[3]=descriptorIndex2;
          indices[4]=0;
          indices[5]=0;
          dummy=multiplier*hh[map62[term]].coefficient[index][neighbour];
          collect_term(indices,dummy,h_id);
        }
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}  


// term21 is term13 backwards
double term13(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
 /*
     TERM 13: - F(k) . F(m) . F(m) . S(+)(k) . S(+)(m)
     TERM 21: - F(k) . F(k) . F(m) . S(+)(k) . S(+)(m)
     ToDo: where is this term used?
  */

  int split0[MAX_LSUBM+1], i1;
  int counter;
//  int mm,nnn,temp1,temp2,temp3, temp, i, j;
  int (*conf0)[5];
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int k, l, m;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
//  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  conf0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=config0[k1][m]; // F(k)
	for(m=1;m<5;m++) conf2[1][m]=config0[k2][m]; // F(m)
	for(m=1;m<5;m++) conf3[1][m]=config0[k2][m]; // F(m)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          ((term==13)? test_k_to_m_neighbour(hh,term, conf1[1], conf2[1]):
           /* 21 */    test_m_to_k_neighbour(hh,term, conf1[1], conf2[1]));
	if (!are_they_neighbours) continue;

        for(counter=0,l=1;l<=length;l++)
        if (l != k1 && l != k2 ) {
          counter++;
          conf0[counter][1]=config0[l][1]; // x
          conf0[counter][2]=config0[l][2]; // y
          conf0[counter][3]=config0[l][3]; // z
          conf0[counter][4]=config0[l][4]; // uc
        }
        
        for (i1=1;i1<=counter;i1++) split0[i1]=0; // initiate splitter
        
        for (i1=0;!i1;) {

          conf1_len=1;
          conf2_len=1;
          conf3_len=1;
          
          // now we have next trimary code 000 -> 001 -> 002 -> 010 ...
          i1=split_trimary(conf0,  counter, split0,
                           conf1,  &conf1_len,  // F(k)
                           conf2,  &conf2_len,  // F(m)
                           conf3,  &conf3_len); // F(m)

          descriptorIndex1=findClusterIndex(conf1_len,conf1);
          if (descriptorIndex1 <= 0) continue;
          
          descriptorIndex2=findClusterIndex(conf2_len,conf2);
          if (descriptorIndex2 <= 0) continue;
          
          descriptorIndex3=findClusterIndex(conf3_len,conf3);	      
          if (descriptorIndex3 <= 0) continue;
          
          multiplier=-1.0;
          
          // coeff from \tilde s^z \tilde s^-
          {
            int indices[7]={0};
            double dummy;
            indices[0]=config_index;
            indices[1]=term; /* 13 or 21 */
            indices[2]=descriptorIndex1;
            indices[3]=descriptorIndex2;
            indices[4]=descriptorIndex3;
            indices[5]=0;
            dummy=multiplier*hh[map62[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
          }
        }
      }
    }
  }
  free(conf0);
  free(conf1);
  free(conf2);
  free(conf3);
  return sum;
}

// s^+ s^z 
//  term14 = term6 backwards
//  term15 = term7 backwards

// s^- s^z
//  term16 = term9 backwards
//  term17 = term8 backwards
//  term18 = term12 backwards
//  term19 = term11 backwards
//  term20 = term10 backwards
//  term21 = term13 backwards

// +-
//  term24 = term22 backwards
// kago_j1j2 (changed JS 0902, to_much_flips)
double term22(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term) {
  /*
    TERM 22: -2 * F(m) . S(+)(k) . S(z)(m)
    TERM 24: -2 * F(k) . S(+)(m) . S(z)(k)     
  */

  int i,j,k,l,m;
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
//  int maxNumberFlips;
//  int numberFlips;
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=MAX_LSUBM+1;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
      
  // (length-1)!        -- permutation factor as we sum over k
  // length             -- internal factor from F(). 

  for(k=1;k<=length;k++) {
    
    index=config0[k][4];
    
    for(counter=0,l=1;l<=length;l++)
    if (l != k) {
      counter++;
      for(m=1;m<5;m++) conf[counter][m]=config0[l][m];
    }

    for(neighbour=1;
        neighbour<=((term==22)?hh[map62[term]].k_to_m_number[index]:
                    /* 24 */   hh[map62[term]].m_to_k_number[index]);
	neighbour++) {
      
      counter=length-1;
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      if (term==22) {  /* k ---> m */
        nn[1]=config0[k][1]+hh[map62[term]].k_to_m[index][neighbour][0];
        nn[2]=config0[k][2]+hh[map62[term]].k_to_m[index][neighbour][1];
        nn[3]=config0[k][3]+hh[map62[term]].k_to_m[index][neighbour][2];
        nn[4]=              hh[map62[term]].k_to_m[index][neighbour][3];
      } else { /* term==24  m ---> k */
        nn[1]=config0[k][1]+hh[map62[term]].m_to_k[index][neighbour][0];
        nn[2]=config0[k][2]+hh[map62[term]].m_to_k[index][neighbour][1];
        nn[3]=config0[k][3]+hh[map62[term]].m_to_k[index][neighbour][2];
        nn[4]=              hh[map62[term]].m_to_k[index][neighbour][3];
      }
      
      counter++;
      conf[counter][1]=nn[1];
      conf[counter][2]=nn[2];
      conf[counter][3]=nn[3];
      conf[counter][4]=nn[4];
      
      // JS0902: this was buggy? config0 was used, but conf should used
      if (to_much_flips(conf,counter,conf[counter]))
          continue;
      
      descriptorIndex=findClusterIndex(counter,conf);
      if (descriptorIndex<=0) continue;
      
      temp_multiplier=UnitCell.s2[nn[4]];
      
      // coeff from \tilde s^+ \tilde s^-
      
      {
        int indices[7]={0};
        double dummy;
        indices[0]=config_index;
        indices[1]=term;
        indices[2]=descriptorIndex;
        indices[3]=0;
        indices[4]=0;
        indices[5]=0;
        dummy=temp_multiplier*hh[map62[term]].coefficient[index][neighbour];
        collect_term(indices,dummy,h_id);
      }
    }
  }
  free(conf);
  return sum;
}  

double term23(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term) {
  /*
   TERM 23: - F(m) . F(m) . S(+)(k) . S(+)(m).
   TERM 25: - F(k) . F(k) . S(+)(k) . S(+)(m).     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int j, k, l, m;
  int k1;
  int k2;
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int tempString[5];
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=config0[k1][m]; // F(m)
	for(m=1;m<5;m++) conf2[1][m]=config0[k1][m]; // F(m)
	for(m=1;m<5;m++) tempString[m]=config0[k2][m]; // s(+)(k)
	
	// loop over nearest neighbours
	
	index=conf1[1][4];      
	
        are_they_neighbours=
           ((term==23)?test_m_to_k_neighbour(hh, term, conf1[1], tempString):
            /* 25 */   test_k_to_m_neighbour(hh, term, conf1[1], tempString));

	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (!are_they_neighbours) continue;
        {
	  
	  for(counter=0,l=1;l<=length;l++)
	  if (l != k1 && l != k2 ) {
	    counter++;
	    tempArray[counter]=l;
	  }
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;      temp=tempArray[j];
		for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
	      } else {
		conf2_len++;      temp=tempArray[j];
		for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findClusterIndex(conf1_len,conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findClusterIndex(conf2_len,conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    multiplier=-1.0;
	    
	    // coeff from \tilde s^+ \tilde s^-
	    {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[map62[term]].coefficient[index][are_they_neighbours];
                collect_term(indices,dummy,h_id);
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

// term24 = term22 backwards
// term25 = term23 backwards

double term26(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 26: S(+)(k) . S(+)(m) 
   */
  
  int tempConfig1[5];
  int tempConfig2[5];
  int k,m;
  int are_they_neighbours;
  int index;
  double sum;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  sum=0.0000;

  if (length != 2) return 0;

  for(k=1;k<=2;k++) {
    
    if (k==1) {
      for(m=1;m<5;m++) tempConfig1[m]=string[1][m]; 
      for(m=1;m<5;m++) tempConfig2[m]=string[2][m]; 
    } else {
      for(m=1;m<5;m++) tempConfig1[m]=string[2][m]; 
      for(m=1;m<5;m++) tempConfig2[m]=string[1][m]; 
    }	

    // loop over nearest neighbours
  
    index=tempConfig1[4];      
    are_they_neighbours=
      test_k_to_m_neighbour(hh, term, tempConfig1, tempConfig2);
    if (!are_they_neighbours) continue;
    
    sum+=1.0*hh[map62[term]].coefficient[index][are_they_neighbours];
  }
  if (sum) {
    int indices[7]={0};
    double dummy;
    indices[0]=config_index;
    indices[1]=term;
    indices[2]=0;
    indices[3]=0;
    indices[4]=0;
    indices[5]=0;
    dummy=sum;
    collect_term(indices,dummy,h_id);
  }
  return sum;
}

double term27(int config_index, int length, int (*string)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
    TERM 27: 4 * H(k)(m) . S(z)(k) . S(z)(m)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'
   */

  int i, j;
  int descriptorIndex;
  double multiplier;
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int x,y,z,cell;
  int index;

//  int numberFlips, maxNumberFlips;

  int (*config0)[5];

  // term27 generates clusters of length+1+1, so we can skip some work
  // for length+1+1 > Params.n we always have an invalid index
  if (length+1+1 > Params.n) return 0;

  config0=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      config0[i][j]=string[i][j];

  for(x=-minX;x<=maxX;x++)  // ToDo: better (faster) x = maxX-m ... m ???
  for(y=-minY;y<=maxY;y++) 
  for(z=-minZ;z<=maxZ;z++) 
  for(cell=1;cell<=UnitCell.number;cell++) {

    // Joerg ToDo: lot of x,y,z pairs lying outside the bounding triangle
    //      and scipped if descriptor not exist, can be done earlier?
    //      same for term30
    
    config0[length+1][1]=x;
    config0[length+1][2]=y;
    config0[length+1][3]=z;
    config0[length+1][4]=cell;
    
    index=cell;

    if (to_much_flips(config0,length+1,config0[length+1]))
       continue;
      
    for(j=1;j<=hh[map62[term]].k_to_m_number[index];j++) {

      // nearest neighbour is at x,y,z, temp put information
      // into configuration and exit.
      
      /* k --> m */
      config0[length+2][1]= x + hh[map62[term]].k_to_m[index][j][0];
      config0[length+2][2]= y + hh[map62[term]].k_to_m[index][j][1];
      config0[length+2][3]= z + hh[map62[term]].k_to_m[index][j][2];
      config0[length+2][4]=     hh[map62[term]].k_to_m[index][j][3];

      if (config0[length+2][1]>maxX || config0[length+2][1]<-maxX) continue;
      if (config0[length+2][2]>maxY || config0[length+2][2]<-maxY) continue;
      if (config0[length+2][3]>maxZ || config0[length+2][3]<-maxZ) continue;

      if (to_much_flips(config0,length+2,config0[length+2])) continue;
      
      multiplier = UnitCell.s2[ config0[length+1][4] ]
                 * UnitCell.s2[ config0[length+2][4] ];
      
      descriptorIndex=findClusterIndex(length+2, config0);
      if (descriptorIndex <= 0) continue;

      // may be invalid if outside bounding box

      // for length+1+1 > Params.n we always have an invalid index

      // and must be exist as config
      {
        int indices[7]={0};
        double dummy;
        indices[0]=config_index;
        indices[1]=term;
        indices[2]=descriptorIndex; // H(k,m) = nn-pair added
        indices[3]=0;
        indices[4]=0;
        indices[5]=0;
        dummy = multiplier * hh[map62[term]].coefficient[index][j];
        collect_term(indices,dummy,h_id);
      }
    }
  }
  free(config0);
  return sum;
}

double term28(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
    TERM 28: 2 * H(k)(m) . H(k)(m) . S(+)(k) . S(+)(m).
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int j, k, l, m;
  int k1;
  int k2;
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=config0[k1][m]; // (k)
	for(m=1;m<5;m++) conf1[2][m]=config0[k2][m]; // (k)
	for(m=1;m<5;m++) conf2[1][m]=config0[k1][m]; // (m)
	for(m=1;m<5;m++) conf2[2][m]=config0[k2][m]; // (m)
	
	// loop over nearest neighbours
	
	index=conf1[1][4];      
	are_they_neighbours=
          test_k_to_m_neighbour(hh,term, conf1[1], conf1[2]);
	if (!are_they_neighbours) continue;

        for(counter=0,l=1;l<=length;l++)
        if (l != k1 && l != k2 ) {
          counter++;
          tempArray[counter]=l;
        }
        
        for(l=0;l<binary_max[counter];l++) {
          
          binary(l,counter,spin);
          conf1_len=2;
          conf2_len=2;
          
          for (j=1;j<=counter;j++) { 
            if (spin[j]) {
              conf1_len++;      temp=tempArray[j];
              for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
            } else {
              conf2_len++;      temp=tempArray[j];
              for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
            }
          }	      
          
          descriptorIndex1=findClusterIndex(conf1_len,conf1);
          if (descriptorIndex1 <= 0) continue;
          
          descriptorIndex2=findClusterIndex(conf2_len,conf2);
          if (descriptorIndex2 <= 0) continue;
          
          multiplier=2.0;
          
          // from \tilde s^- \tilde s^-
          {
            int indices[7]={0};
            double dummy;
            indices[0]=config_index;
            indices[1]=term;
            indices[2]=descriptorIndex1;
            indices[3]=descriptorIndex2;
            indices[4]=0;
            indices[5]=0;
            dummy=multiplier*
              hh[map62[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
          }
        }
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

double term29(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term) {
  /*
    TERM 29: 4 * H(k)(m) . F(m) . S(+)(m) . S(z)(k)
    TERM 35: 4 * H(k)(m) . F(k) . S(+)(k) . S(z)(m)
     
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int index;
  int nn[5];
  int neighbour;
//  int maxNumberFlips;
//  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
//  int temp_number_conf2_sites;
  double sum;

  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  }

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=config0[k][m]; // F(m)
      conf2[1][m]=config0[k][m]; // H(m,k) <-- m here.
    }
      
    for(counter=0,m=1;m<=length;m++)
    if (m != k) {
	counter++;
	tempArray[counter]=m;
    }    
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {      // F(m)
	  conf1_len++;      temp=tempArray[j];
	  for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
	} else {            // H(k,m)
	  conf2_len++;      temp=tempArray[j];
	  for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
	}
      }
      
      index=conf2[1][4];
 
      for(neighbour=1;
          neighbour<=((term==29)?hh[map62[term]].m_to_k_number[index]:
                      /* 35 */   hh[map62[term]].k_to_m_number[index]);
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
        if (term==29) {
  	  /* m ---> k */
	  nn[1]=conf2[1][1]+hh[map62[term]].m_to_k[index][neighbour][0];
	  nn[2]=conf2[1][2]+hh[map62[term]].m_to_k[index][neighbour][1];
	  nn[3]=conf2[1][3]+hh[map62[term]].m_to_k[index][neighbour][2];
	  nn[4]=            hh[map62[term]].m_to_k[index][neighbour][3];
	} else {
  	  /* k ---> m */
	  nn[1]=conf2[1][1]+hh[map62[term]].k_to_m[index][neighbour][0];
	  nn[2]=conf2[1][2]+hh[map62[term]].k_to_m[index][neighbour][1];
	  nn[3]=conf2[1][3]+hh[map62[term]].k_to_m[index][neighbour][2];
	  nn[4]=            hh[map62[term]].k_to_m[index][neighbour][3];
        }

        conf2[conf2_len+1][1]=nn[1];
        conf2[conf2_len+1][2]=nn[2];
        conf2[conf2_len+1][3]=nn[3];
        conf2[conf2_len+1][4]=nn[4];
        
        if (to_much_flips(conf2, conf2_len+1, conf2[conf2_len+1]))
           continue;

        descriptorIndex1=findClusterIndex(conf1_len,conf1);
        if (descriptorIndex1 <= 0) continue;
        
        descriptorIndex2=findClusterIndex(conf2_len+1,conf2);
        if (descriptorIndex2 <= 0) continue;
        
        multiplier=1.0;

        multiplier=(-4)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- k
        
        // coefficient from \tilde s^- \tilde s^-
        {
          int indices[7]={0};
          double dummy;
          indices[0]=config_index;
          indices[1]=term;
          indices[2]=descriptorIndex1;
          indices[3]=descriptorIndex2;
          indices[4]=0;
          indices[5]=0;
          dummy=multiplier*
            hh[map62[term]].coefficient[index][neighbour];
          collect_term(indices,dummy,h_id);
        }
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

/* this is the slowest term, 75%..90%(Cubic) of CPU (Okt2004, Joerg) */
double term30(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{

  /*
     TERM 30: 4 * F(k) . F(m) . S(z)(k) . S(z)(m)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int j, k, l, m;
  double multiplier;

  int x,y,z,cell;
  int index;

//  int numberFlips, maxNumberFlips;

  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int counter;
  int spin[MAX_LSUBM+2];
  int neighbour;
  int temp_x;
  int temp_y;
  int temp_z;
  int temp_cell;
  int descriptorIndex1;
  int descriptorIndex2;
  double sum;
  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(x=-minX;x<=maxX;x++)
  for(y=-minY;y<=maxY;y++)
  for(z=-minZ;z<=maxZ;z++)
  for(cell=1;cell<=UnitCell.number;cell++) {
    
    // ToDo: skip x+y+z outside triangle range [0..m]?

    index=cell;

    counter=length;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);   /* convert l to binary string spin[] */

      // spin[j]: 0 = F(m)  1 = F(k)
      // copy sites for 0-bit to conf2 and sites for 1-bit to conf1
      conf1_len=0;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
        if (spin[j]) {
          conf1_len++;
          for (m=1;m<5;m++) conf1[conf1_len][m]=config0[j][m];
        } else {
          conf2_len++;
          for (m=1;m<5;m++) conf2[conf2_len][m]=config0[j][m];
        }
      }
      
      // Took this outside inner loop (Joerg Sep2005)
      conf1[conf1_len+1][1]=x;
      conf1[conf1_len+1][2]=y;
      conf1[conf1_len+1][3]=z;
      conf1[conf1_len+1][4]=cell;
   
      // ToDo: better split first and and add neighbour
      //   or add neighbour first?

      if (to_much_flips(conf1, conf1_len+1, conf1[conf1_len+1]))
          continue;

      descriptorIndex1=findClusterIndex(conf1_len+1,conf1);
      if (descriptorIndex1 <= 0) continue;
        
      for(neighbour=1;
          neighbour<=hh[map62[term]].k_to_m_number[cell];
          neighbour++) {
        
        // nearest neighbour is at x,y,z, cell put information
        // into configuration and exit.
        
        /* k ---> m */
        
        temp_x = x + hh[map62[term]].k_to_m[cell][neighbour][0];
        temp_y = y + hh[map62[term]].k_to_m[cell][neighbour][1];
        temp_z = z + hh[map62[term]].k_to_m[cell][neighbour][2];
        temp_cell=   hh[map62[term]].k_to_m[cell][neighbour][3];
        
        conf2[conf2_len+1][1]=temp_x;
        conf2[conf2_len+1][2]=temp_y;
        conf2[conf2_len+1][3]=temp_z;
        conf2[conf2_len+1][4]=temp_cell;

        if (temp_x>maxX || temp_x<-maxX) continue; // JS090322
        if (temp_y>maxY || temp_y<-maxY) continue;
        if (temp_z>maxZ || temp_z<-maxZ) continue;

        if (to_much_flips(conf2, conf2_len+1, conf2[conf2_len+1]))
            continue;

        multiplier=UnitCell.s2[     cell]*
                   UnitCell.s2[temp_cell];
        
        // coeff from \tilde s^- \tilde s^- 
        
        descriptorIndex2=findClusterIndex(conf2_len+1,conf2);
        if (descriptorIndex2 <= 0) continue;
        
        {
          int indices[7]={0};
          double dummy;
          indices[0]=config_index;
          indices[1]=term;
          indices[2]=descriptorIndex1;
          indices[3]=descriptorIndex2;
          indices[4]=0;
          indices[5]=0;
          dummy=multiplier
               *hh[map62[term]].coefficient[index][neighbour];
          collect_term(indices,dummy,h_id);
        }
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

// term33 is term31 backwards
double term31(int config_index, int conf0_len, int (*conf0)[5],
              struct hamiltonian_defn hh[], int term)
{
  
  /*
    TERM 31: 2 * F(m) . F(m) . F(k) . S(+)(m) . S(z)(k)  (like term33)
    TERM 33: 2 * F(k) . F(k) . F(m) . S(+)(k) . S(z)(m)  (like term31)

   split cfg using !k into 3cfgs, add k to cfg 1+2, add H?? to cfg3
   see SrCuBO-4-4 cfg6 to 6,3,1  3,6,1  4,5,1  5,4,1  4,3,2  3,4,2 
       len          3     3 1 1  1 3 1  2 2 1  2 2 1  2 1 2  1 2 2
       len(-1,-1,-0)      2 0 1  0 2 1  1 1 1  1 1 1  1 0 2  0 1 2  
    6= (0,0,0,4) (0,0,0,3) (0,0,0,2) \|
    5= (0,0,0,3) (0,0,0,2) |
    4= (0,0,0,4) (0,0,0,2) \
    3= (0,0,0,1)           .
    2= (0,0,0,4) (0,0,0,3) -
    1= (0,0,0,3)           '    # == (0,0,0,1)? missing sym?
   see SrCuBO-2-2 cfg3 to 3,3,1 l=1 split nothing, but cfg1=k,cfg2=k,cfg3=m
       len         1      1 1 1
       len(-1,-1,-0)      0 0 1
    3= (0,0,3)
    1= (0,0,1)
    sum to 0 for alpha=0 ?!!!
     
  */
  
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  // int spin[MAX_LSUBM+1];
  static int split1[MAX_LSUBM+1];
  static int split2[MAX_LSUBM+1];
  int counter;
  int len2;
  // int temp;
  int (*conf1)[5], (*conf2)[5], (*conf3)[5], (*rest1)[5], (*rest2)[5];
  int  j, k, l, m;
  int nn[5];
//  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
//  int tempArray[MAX_LSUBM+1];
//  int temp_number_conf3_sites;
//  int numberFlips;
//  int maxNumberFlips;
  double sum;
  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k=1;k<=conf0_len;k++) {
    
    for(m=1;m<5;m++) conf1[1][m]=conf0[k][m]; // F(k)
    for(m=1;m<5;m++) conf2[1][m]=conf0[k][m]; // F(k)
    
    // rest1 is conf0 without site k
    
    for(counter=0,l=1;l<=conf0_len;l++)
    if (l != k) {
        counter++;
        for(m=1;m<5;m++) rest1[counter][m]=conf0[l][m];
    } // counter is conf0_len-1
/*
   see SrCuBO-2-2 cfg3 to 3,3,1
       len         1      1 1 1
       len(-1,-1,-0)      0 0 1
    3= (0,0,3) len=1
    1= (0,0,1)
*/   
    /* decompose the rest spins to 3 configs (stepwise 2 times to 2 configs) */
    { int i; for (i=1;i<=conf0_len-1;i++) split1[i]=0; }
    int i1=0; // split counter overflow
    for(;;) {
      conf1_len=1; // first site is k before split0 is appended
      len2=0;      // rest is empty   before split1 is appended
      if (i1) break; 
      /* split configuration and count up (new: no index list) */
      i1= split_binary(rest1, conf0_len-1, split1,
                       conf1, &conf1_len,
                       rest2, &len2);

      /* speedup Apr06 JS */
//      if (!subsetClusters(conf1,conf1_len)) continue;
      // moved outside inner loop, Joerg Sep2005, speedup
      descriptorIndex1=findClusterIndex(conf1_len, conf1);
      if (descriptorIndex1 <= 0) continue;
      
#if 0
  if (Control.output > 8){
   printf("Term33 idx= %d len=%d k=%d\n",config_index,conf0_len,k);
   for( int i=1; i<=conf0_len; i++)
     printf(" %d x y uc %d %d %d\n",i,conf0[i][1], conf0[i][2], conf0[i][4]);
   printf(" conf1  len=%d\n", conf1_len);
   for( int i=1; i<=conf1_len; i++)
     printf(" %d x y uc %d %d %d\n",i,conf1[i][1], conf1[i][2], conf1[i][4]);
   printf(" rest2  len=%d\n", len2);
   for( int i=1; i<=len2; i++)
     printf(" %d x y uc %d %d %d\n",i,rest2[i][1], rest2[i][2], rest2[i][4]);
 }
#endif

      { int i; for (i=1;i<=len2;i++) split2[i]=0; }
      int i2=0; // split counter overflow
      for(;;) {
        conf2_len=1; // first site is k before split0 is appended
        conf3_len=0; // empty           before split1 is appended
        if (i2) break;
        /* split configuration and count up (new: no index list) */
        i2= split_binary(rest2, len2, split2,
                         conf2, &conf2_len,
                         conf3, &conf3_len);
	  
//        if (!subsetClusters(conf2,conf2_len)) continue;
        descriptorIndex2=findClusterIndex(conf2_len, conf2);
        if (descriptorIndex2 <= 0) continue;

#if 0
  if (Control.output > 8){
   printf(" conf2  len=%d\n", conf2_len);
   for( int i=1; i<=conf2_len; i++)
     printf(" %d x y uc %d %d %d\n",i,conf2[i][1], conf2[i][2], conf2[i][4]);
   printf(" conf3  len=%d\n", conf3_len);
   for( int i=1; i<=conf3_len; i++)
     printf(" %d x y uc %d %d %d\n",i,conf3[i][1], conf3[i][2], conf3[i][4]);
  }
#endif
//	are_they_neighbours=0;
	index=conf1[1][4];   /* uc of site 1 */
        /* ToDo: make a subroutine! */
	for(j=1;j<=((term==31)?hh[map62[term]].m_to_k_number[index]:
                               hh[map62[term]].k_to_m_number[index]);j++) {
	    
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
          if (term==31) {
  	    /* m ---> k */
	    nn[1]=conf1[1][1]+hh[map62[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map62[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map62[term]].m_to_k[index][j][2];
	    nn[4]=            hh[map62[term]].m_to_k[index][j][3];
          } else { // term33
  	    /* k ---> m */
	    nn[1]=conf1[1][1]+hh[map62[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map62[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map62[term]].k_to_m[index][j][2];
	    nn[4]=            hh[map62[term]].k_to_m[index][j][3];
          }

          // append k --> m
          // temp_number_conf3_sites=conf3_len+1;
          conf3[conf3_len+1][1]=nn[1];
          conf3[conf3_len+1][2]=nn[2];
          conf3[conf3_len+1][3]=nn[3];
          conf3[conf3_len+1][4]=nn[4];
          if (to_much_flips(conf3,conf3_len+1,conf3[conf3_len+1]))
             continue;

          { 
#if 0
  if (Control.output > 8)
   printf(" conf3  len=%d OK1\n", conf3_len+1);
#endif	      
	      descriptorIndex3=findClusterIndex(conf3_len+1, conf3);
              if (descriptorIndex3 <= 0) continue;
#if 0     
  if (Control.output > 8)
   printf(" conf3  len=%d OK2\n", conf3_len+1);
#endif
	      multiplier=1.0;
	      
	      // multiplier*=conf1_len;    // F(k)
	      // multiplier*=conf2_len;    // F(k)
	      // multiplier*=conf3_len+1;  // F(m)
	      
	      multiplier=(-2)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- m
	      
	      // from \tilde s^- \tilde s^-
	      
	      {
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=term; /* 31 or 33 */
		  indices[2]=descriptorIndex1;
		  indices[3]=descriptorIndex2;
		  indices[4]=descriptorIndex3;
		  indices[5]=0;
		  dummy=multiplier*hh[map62[term]].coefficient[index][j];
                  collect_term(indices,dummy,h_id);
	      }
	      
	    }
	  }
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(rest1);
  free(rest2);
  return sum;
}

double term32(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /* 
     TERM 32: 4 * F(k) . F(m) . H(k)(m) . S(+)(k) . S(+)(m)
   */
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int split0[MAX_LSUBM+1];
  int counter, i1;
//  int temp, i, j;
  int (*conf0)[5];
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int k, l, m;
  int k1;
  int k2;
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  double sum;
  sum=0.0000;

  conf0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++)
  for(k2=1;k2<=length;k2++)
  if (k2 != k1) {
    
    for(m=1;m<5;m++) conf1[1][m]=config0[k1][m]; // F(k)
    for(m=1;m<5;m++) conf2[1][m]=config0[k2][m]; // F(m)
    for(m=1;m<5;m++) conf3[1][m]=config0[k1][m]; // H(k,m) <--- k
    for(m=1;m<5;m++) conf3[2][m]=config0[k2][m]; // H(k,m) <--- m
    
    index=conf1[1][4];      

    are_they_neighbours =
      test_k_to_m_neighbour(hh, term, conf1[1], conf2[1]);
    
    // conf1[1] and conf2[1] need to be nearest-neighbours 
    
    if (!are_they_neighbours) continue;
    
    for(counter=0,l=1;l<=length;l++) 
    if (l != k1 && l != k2 ) {
      counter++;
      conf0[counter][1]=config0[l][1]; // x
      conf0[counter][2]=config0[l][2]; // y
      conf0[counter][3]=config0[l][3]; // z
      conf0[counter][4]=config0[l][4]; // uc
    }
    
    for (i1=1;i1<=counter;i1++) split0[i1]=0; // initiate splitter

    for (i1=0;!i1;) {
      
      conf1_len=1;
      conf2_len=1;
      conf3_len=2;
       
      i1=split_trimary(conf0,  counter, split0,
                       conf1,  &conf1_len,
                       conf2,  &conf2_len,
                       conf3,  &conf3_len);
        
      descriptorIndex1=findClusterIndex(conf1_len,conf1);
      if (descriptorIndex1 <= 0) continue;

      descriptorIndex2=findClusterIndex(conf2_len,conf2);
      if (descriptorIndex2 <= 0) continue;

      descriptorIndex3=findClusterIndex(conf3_len,conf3);
      if (descriptorIndex3 <= 0) continue;

      // from \tilde s^- \tilde s^-
      {
        int indices[7]={0};
        double dummy;
        indices[0]=config_index;
        indices[1]=term;
        indices[2]=descriptorIndex1;
        indices[3]=descriptorIndex2;
        indices[4]=descriptorIndex3;
        indices[5]=0;
        dummy=4.0*
          hh[map62[term]].coefficient[index][are_they_neighbours];
        collect_term(indices,dummy,h_id);
      }
    }
  }
  free(conf0);
  free(conf1);
  free(conf2);
  free(conf3);
  return sum;
}

// term33 is term31 backwards

/* slowest 31% for chain-6-6 Joerg Sep05  new algo */
/* length = 1..MAX_LSUBM  */
double term34(int config_index, int length, int (*config)[5], struct hamiltonian_defn hh[], int term)
{

  /*
     
     TERM 34: 09.34  F(k) . F(k) . F(m) . F(m) . S(+)(k) . S(+)(m)
   
    */

  // < cfg0 | H | cfg1 cfg2 cfg3 cfg4 (+2fips) >
  // length+2 >= 4*minlen = 4
  // ToDo: max(minlen,len0+2-3maxlen) <= lencfg1 <= len0+2-3minlen
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  static int split1[MAX_LSUBM+1]; /* decompose bit pattern 1=take, 0=rest */
  static int split2[MAX_LSUBM+1]; /* decompose bit pattern 1=take, 0=rest */
  static int split3[MAX_LSUBM+1]; /* decompose bit pattern 1=take, 0=rest */
  int len1, len2, len3;
  // int temp;
  int (*conf[4])[5]; // Joerg: replaces old (*conf1)[5]..**conf4
  int (*rest1)[5], (*rest2)[5], (*rest3)[5];
  int k, l, m, mm;
  int k1, k2, ks;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex[4]; // Joerg: replace old descriptor1..4
  // double multiplier;
  int conf_len[4]; // Joerg: replaces conf[1..4]_len
  double sum;
  sum=0.0000;

  /* config: {{unused,},x,y,z,uc} */

  conf[0]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));   // default: MAX_LSUB=20
  conf[1]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf[2]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf[3]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest1  =(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest2  =(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest3  =(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  // ToDo: not needed? remove
  for(k=1;k<=MAX_LSUBM;k++)
   for(l=1;l<5;l++) {
      conf[0][k][l]=0;
      conf[1][k][l]=0;
      conf[2][k][l]=0;
      conf[3][k][l]=0;
  } 
 
  // pick up 2 flips k1,k2, which are k_to_m_neighbours
  for(k1=1;k1<=length;k1++)
   for(k2=1;k2<=length;k2++)
    if (k1 != k2) { // old: if (k2 != k1)  Sep05, use 12 <-> 34 symmetry
    // test: if (k1 < k2) {} // old: if (k2 != k1)  Sep05, use 12 <-> 34 symmetry
    // Buggy for SrCuO4-2-2 + Alpha (2 09.34 1 1 1 1 1.0) missing
    //   should be ok, if also for m-to-k neighbourhood is checked !?
    // k1 and k2 must fullfill restrictions! k1 from A than k2 from B-sublat.!
	
	for(m=1;m<5;m++) {
	  conf[0][1][m]=config[k1][m]; // F(k) 12 and 34 are symmetric!
	  conf[1][1][m]=config[k1][m]; // F(k)
	  conf[2][1][m]=config[k2][m]; // F(m)
	  conf[3][1][m]=config[k2][m]; // F(m)
        }
	
	// loop over nearest neighbours
	
        index = conf[0][1][4];
        are_they_neighbours=
           test_k_to_m_neighbour(hh,term,conf[0][1],conf[2][1]);
	// conf[0][1] and conf[2][1] need to be nearest-neighbours 
        //  == config[k1] and config[k2] (check this earlier!)
        // return: 0=no_neighbours, 1..n=neighbour_index
	
	if (!are_they_neighbours) continue;
        
        // remove sites k1 and k2 and prepare rest for decomposition
        len1=0;
        for(l=1;l<=length;l++) {     // k3
          if (l != k1 && l != k2) {  // different from k1 and k2
            len1++;
            for (mm=1;mm<5;mm++) rest1[len1][mm] = config[l][mm];
          }
        }
        
        // now we decompose configuration into 4 configurations
        //  ToDo: better do successiv decomposition into 2 configurations?
        //        than decompose 2nd config again into 2 parts
        //        1. decompose tmp  binary to Config(1) + rest
        //        2. decompose rest binary to Config(2) + tmp
        //        3. decompose tmp  binary to Config(3) + Config(4)
        // (quartery) quartary code decomposed to dual + trimary code
        // 000 001 002 003 010 011 012 => 000 001 010 011 100 101
        //                                000 00  0 0
        //                                001 01  0 1
        //                                002 02  0 2
        //                                010 10  1 0
        // ToDo: further speedup by filter out len1<=len0+2-3*minlen
  // ToDo: as function? config0 auf config1..n aufteilen (anhaengen)
 {
        { int i; for (i=1;i<=len1;i++) split1[i]=0; }
        int i1=0; // counter overflow
        for(;;) {
          conf_len[0]=1; // first site is k1 or k2
          len2=0;
          if (i1) break;
          /* split configuration and count up (new: no index list) */
          i1= split_binary(rest1,   len1, split1,
                           conf[0], &conf_len[0],
                           rest2,   &len2);
          ks=0; 
          // this consumes most of CPU time
          //  square_s1-8-4 configs=30 calls=604004/1508970
          //    decompositions: cfgs^4=30^4=810e3
          // new Mar06 testwise (speed*2)
//          if (!subsetClusters(conf[ks],conf_len[ks])) continue;
          // ToDo: return descrIndex from descriptorIndex=findClusterIndex?
          descriptorIndex[ks]=findClusterIndex(conf_len[ks],conf[ks]);
          // output here testwise???
          if (descriptorIndex[ks] <= 0) continue; // mostly invalid
            
        /* decompose the rest spins to other 3 configs */
        { int i; for (i=1;i<=len2;i++) split2[i]=0; }
        int i2=0; // counter overflow
        for(;;) {
          conf_len[1]=1; // first site is k1 or k2 before split appended
          len3=0;
          if (i2) break;
          /* split configuration and count up (new: no index list) */
          i2= split_binary(rest2,   len2, split2,
                           conf[1], &conf_len[1],
                           rest3,   &len3);
          ks=1; 
//          if (!subsetClusters(conf[ks],conf_len[ks])) continue;
          descriptorIndex[ks]=findClusterIndex(conf_len[ks],conf[ks]);
          if (descriptorIndex[ks] <= 0) continue; // mostly invalid
            
        /* decompose the rest spins to other 2 configs */
        { int i; for (i=1;i<=len3;i++) split3[i]=0; }
        int i3=0; // counter overflow
        for(;;) {
          conf_len[2]=1; // first site is k1 or k2
          conf_len[3]=1;
          if (i3) break;
          /* split configuration and count up (new: no index list) */
          i3=  split_binary(rest3,   len3, split3,
                            conf[2], &conf_len[2],
                            conf[3], &conf_len[3]);
              //  check if we have 4 valid configs conf[] here
              for (ks=2; ks<4; ks++) {
//                if (!subsetClusters(conf[ks],conf_len[ks])) break;
                descriptorIndex[ks]=findClusterIndex(conf_len[ks],conf[ks]);
                if (descriptorIndex[ks] <= 0) break; // mostly invalid
              }

              // Joerg ToDo: skip     if descrIndex[2]>descrIndex[1]
              //             factor 2 if descrIndex[2]<descrIndex[1]
              if (ks<4) continue; // skip the rest

              // Joerg: now we have 4 valid indices to configs
              //        its probably faster to test NumConfigs^4 (<=length-1)
              //        if they build our config (length, one same site + miss side)
              //  decompose into 4 configs (2 * 2!2!= (k1,k2) * (k1,k2)orders) 
              // LC-4-4 no term
              // LC-6-6 only Config9: 16*(5,2,2,1)+24*(2,2,1,1)
              // oooooo -> 16 * o....o o..o o..o oo + 24 * o..o o..o oo oo
              //  sum_len(cfg1..4)=length(conf9)+2 ???
              //  first two configs and last two configs are symmetric!
              //        -> always factor 8 (divide to 12 and 34 and fact=2)
              //  we test n^2*4^(n-2) decompositions here (grows exponentially)
              //   

              // multiplier=1.0;
              // multiplier = * (conf_len[0]-1)!
              //              * (conf_len[1]-1)!
              //              * (conf_len[2]-1)!
              //              * (conf_len[3]-1)!
              // in each case:
              // (length-1)!      -- we sum over k1 and k2.
              // length           -- internal factor in F().
              
              {
                  static int indices[7]={0};
                  double dummy;
                  indices[0]=config_index;  // original config
                  indices[1]=term;
                  indices[2]=descriptorIndex[0]; // modified config1..4
                  indices[3]=descriptorIndex[1];
                  indices[4]=descriptorIndex[2];
                  indices[5]=descriptorIndex[3];
                  dummy= // 2*  // factor 2 for 3412 sym, Joerg Sep05
                   hh[map62[term]].coefficient[index][are_they_neighbours];
                  collect_term(indices,dummy,h_id);
              }
              
            }
          }
        }
        }
  } // k1!=k2
  free(conf[0]);
  free(conf[1]);
  free(conf[2]);
  free(conf[3]);
  free(rest1  );
  free(rest2  );
  free(rest3  );
  return sum;  // not used!
}

// term35 = term29 backwards
// now the ``extra'' terms for s > 1/2.
double term36(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 36: 2 * M(k)(k)(m) . S(+)(k) . S(z)(m)
     TERM 37: 2 * M(k)(m)(m) . S(+)(m) . S(z)(k)
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptorIndex;
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  conf   =(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=MAX_LSUBM;i++) for(j=1;j<5;j++) conf[i][j]=0;

  for(k=1;k<=length;k++) {
    
    index=config0[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips <= 1) continue;
    
    for(j=1;j<=length;j++)
      for(m=1;m<5;m++) conf[j][m]=config0[j][m]; // M(k,k,m)
    
    // we repeat the `k' 
    
    for(m=1;m<5;m++) conf[length+1][m]=config0[k][m];

    index=config0[k][4];  // JS0902: missed on TERM 37
    
    for(neighbour=1;neighbour<=
            ((term==36)?hh[map62[term]].k_to_m_number[index]:
              /* 37 */  hh[map62[term]].m_to_k_number[index]);
        neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      
      if (term==36) {
        /* k ---> m */
        nn[1]=config0[k][1]+hh[map62[term]].k_to_m[index][neighbour][0];
        nn[2]=config0[k][2]+hh[map62[term]].k_to_m[index][neighbour][1];
        nn[3]=config0[k][3]+hh[map62[term]].k_to_m[index][neighbour][2];
        nn[4]=                    hh[map62[term]].k_to_m[index][neighbour][3];
      } else { /* 37 */
        /* m ---> k */
        nn[1]=config0[k][1]+hh[map62[term]].m_to_k[index][neighbour][0];
        nn[2]=config0[k][2]+hh[map62[term]].m_to_k[index][neighbour][1];
        nn[3]=config0[k][3]+hh[map62[term]].m_to_k[index][neighbour][2];
        nn[4]=                    hh[map62[term]].m_to_k[index][neighbour][3];
      }

      conf[length+2][1]=nn[1]; // `m' is the nearest neighbour.
      conf[length+2][2]=nn[2];
      conf[length+2][3]=nn[3];
      conf[length+2][4]=nn[4];
      
      numberFlips1=1;
      for(l=1;l<=length+1;l++)
        if (nn[1]==conf[l][1] &&
            nn[2]==conf[l][2] &&
            nn[3]==conf[l][3] &&
            nn[4]==conf[l][4])
          numberFlips1++;
      maxNumberFlips1=UnitCell.s2[nn[4]];
      if (numberFlips1 > maxNumberFlips1 )  continue;
      
      numberFlips2=0;
      for(l=1;l<=length+1;l++)
        if (config0[k][1]==conf[l][1] &&
            config0[k][2]==conf[l][2] &&
            config0[k][3]==conf[l][3] &&
            config0[k][4]==conf[l][4])
          numberFlips2++;
      maxNumberFlips2=UnitCell.s2[index];
      if (numberFlips2 > maxNumberFlips2) continue;
      
      descriptorIndex=findClusterIndex(length+2,conf);
      if (descriptorIndex <= 0) continue;
      
      temp_multiplier=-2*0.5*UnitCell.s2[nn[4]];
      
      // coefficient from \tilde s^- \tilde s^- expansion
      {
        int indices[7]={0};
        double dummy;
        indices[0]=config_index;
        indices[1]=term; /* 36 or 37 */
        indices[2]=descriptorIndex;
        indices[3]=0;
        indices[4]=0;
        indices[5]=0;
        dummy=temp_multiplier*
          hh[map62[term]].coefficient[index][neighbour];
        collect_term(indices,dummy,h_id);
      }
    }
  }
  free(conf);
  return sum;
}

// term37 = term36 backwards
  
double term38(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 38: N(k)(k)(m)(m) . S(+)(k) . S(+)(m).
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int i, j, k, l, m;
  int k1;
  int k2;
  int are_they_neighbours;
  int index, index1, index2;
  int maxNumberFlips1, maxNumberFlips2;
  int numberFlips1, numberFlips2;
  int descriptorIndex;
  // double multiplier;
  int (*conf)[5];
  double sum;
  sum=0.0000;

  conf=  (int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(k=1;k<MAX_LSUBM+3;k++) {
    for(l=1;l<5;l++) {
      conf[k][l]=0;
    }
  } 
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++) conf[i][j]=config0[i][j];
	
	// now add in extra `k' and `m' sites for N(k,k,m,m)
	
	for(m=1;m<5;m++) conf[length+1][m]=config0[k1][m]; // `k'
	for(m=1;m<5;m++) conf[length+2][m]=config0[k2][m]; // `m'
	
	numberFlips1=1;
	numberFlips2=1;
	
	for(l=1;l<=length;l++) {
	  if (config0[k1][1]==conf[l][1] &&
	      config0[k1][2]==conf[l][2] &&
	      config0[k1][3]==conf[l][3] &&
	      config0[k1][4]==conf[l][4])
	    numberFlips1++;
	
	  if (config0[k2][1]==conf[l][1] &&
	      config0[k2][2]==conf[l][2] &&
	      config0[k2][3]==conf[l][3] &&
	      config0[k2][4]==conf[l][4])
	    numberFlips2++;
	}

	index1=config0[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];
	
	index2=config0[k2][4];
	maxNumberFlips2=UnitCell.s2[index2];

	index=config0[k1][4]; 

	are_they_neighbours=
          test_k_to_m_neighbour(hh,term, config0[k1], config0[k2]);

        // ToDo: why is maxNumberFlips* == 1 excluded? melting?
	if ((maxNumberFlips1 > 1) && 
	    (maxNumberFlips2 > 1) &&
	    (numberFlips1 <= maxNumberFlips1) &&
	    (numberFlips2 <= maxNumberFlips2) &&
	    (are_they_neighbours)) {
	  
	  descriptorIndex=findClusterIndex(length+2,conf);
	  
          if (descriptorIndex > 0) 
          {
            int indices[7]={0};
            double dummy;
            indices[0]=config_index;
            indices[1]=term;
            indices[2]=descriptorIndex;
            indices[3]=0;
            indices[4]=0;
            indices[5]=0;
            dummy=
              hh[map62[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
          }
	}
      }
    }
  }
  free(conf);
  return sum;
}

double term39(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 39: F(m) . M(k)(k)(m) . S(+)(k) . S(+)(m) = -- [2]
     TERM 41: F(k) . M(m)(m)(k) . S(+)(k) . S(+)(m) = -- [2]
   */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5], (*conf2)[5];
  int j, k, l, m;
  int k1, k2;
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;

  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {  // have to be neighbour
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[ config0[k1][4] ];
	
	if (maxNumberFlips <= 1) continue; // want to flip site k=k1 twice

        for(m=1;m<5;m++) {
          conf1[1][m]=config0[k1][m]; // M(k,k,m) k--1st one
          conf1[2][m]=config0[k1][m]; // M(k,k,m) k--2nd one
          conf1[3][m]=config0[k2][m]; // M(k,k,m) m 
          conf2[1][m]=config0[k2][m]; // F(m)     m
        }
        
        // loop over nearest neighbours
        
        index=conf1[1][4];      
        are_they_neighbours=
         ((term==39)? test_k_to_m_neighbour(hh, term, conf1[1], conf2[1]):
          /* 41 */    test_m_to_k_neighbour(hh, term, conf1[1], conf2[1]));

        // conf1[1] and conf2[1] need to be nearest-neighbours 
        
        if (!are_they_neighbours) continue;
        
        // distribute the rest of config0 (length-2) to 2 new configs
        //   cfg1_len=3+r1
        //   cfg2_len=1+r2   r1+r2=len-2
        for(counter=0,l=1;l<=length;l++)
        if (l != k1 && l != k2 ) {
          counter++;
          tempArray[counter]=l;
        }
        
        for(l=0;l<binary_max[counter];l++) {
          
          binary(l,counter,spin);
          conf1_len=3;
          conf2_len=1;
          
          for (j=1;j<=counter;j++) { 
            temp=tempArray[j];
            if (spin[j]) { // H
              conf1_len++;
              for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
            } else {       // H
              conf2_len++;
              for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
            }
          }
          
          // make sure that we don't have too many `k' sites.
          
          numberFlips1=0;
          
          for(m=1;m<=conf1_len;m++) {
            if (config0[k1][1]==conf1[m][1] &&
                config0[k1][2]==conf1[m][2] &&
                config0[k1][3]==conf1[m][3] &&
                config0[k1][4]==conf1[m][4])
              numberFlips1++;
          }

          maxNumberFlips1=UnitCell.s2[ config0[k1][4] ];
          
          if (numberFlips1 > maxNumberFlips1) continue;
          
          descriptorIndex1=findClusterIndex(conf1_len,conf1);
          if (descriptorIndex1 <= 0) continue;
          
          descriptorIndex2=findClusterIndex(conf2_len,conf2);		
          if (descriptorIndex2 <= 0) continue;
          
          multiplier=1.0;

          // for(i=1;i<=(conf1_len-3);i++) multiplier=multiplier*i;
          //  multiplier=multiplier*conf1_len*(conf1_len-1)*(conf1_len-2);

          // M(k,k,m)             
          // (length-3)!                  -- because we sum over k and m
          // length*(length-1)*(length-2) -- factor due to M().
          
          // for(i=1;i<conf2_len;i++) multiplier=multiplier*i;
          // multiplier=multiplier*conf2_len;
          
          // F(m)             
          // (length-1)!        -- because we sum over k and m
          // length             -- factor due to F().

          multiplier=2.0;

          // coefficient from \tilde s^- \tilde s^- expansion
          
          {
            int indices[7]={0};
            double dummy;
            indices[0]=config_index;
            indices[1]=term;
            indices[2]=descriptorIndex1;
            indices[3]=descriptorIndex2;
            indices[4]=0;
            indices[5]=0;
            dummy=multiplier*
              hh[map62[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
          }
        }
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

double term40(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 40:  2 * F(k) . H(m)(m) . S(+)(m) . S(z)(k)
     TERM 42:  2 * F(m) . H(k)(k) . S(+)(k) . S(z)(m)
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  double sum;
  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) { // `m' here. 
    
    index=config0[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(m=1;m<5;m++) {
	conf1[1][m]=config0[k][m]; // H(m,m) -- 1st m 
	conf1[2][m]=config0[k][m]; // H(m,m) -- 2nd m
      }
      
      for(counter=0,m=1;m<=length;m++)
      if (m != k) {
	  counter++;
	  tempArray[counter]=m;
      }
      
      for(l=0;l<binary_max[counter];l++) {
	
	binary(l,counter,spin);
	conf1_len=2;
	conf2_len=0;
	
	for (j=1;j<=counter;j++) { 
	  if (spin[j]) {      // F(m)
	    conf1_len++;      temp=tempArray[j];
	    for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
	  } else {            // H(k,m)
	    conf2_len++;      temp=tempArray[j];
	    for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
	  }
	}
	
	index=conf1[1][4];
	
	for(neighbour=1;
            neighbour<=((term==40)?hh[map62[term]].m_to_k_number[index]:
                          /* 42 */ hh[map62[term]].k_to_m_number[index]);
	    neighbour++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  if (term==40) {
 	    /* m ---> k */
	    nn[1]=conf1[1][1]+hh[map62[term]].m_to_k[index][neighbour][0];
	    nn[2]=conf1[1][2]+hh[map62[term]].m_to_k[index][neighbour][1];
	    nn[3]=conf1[1][3]+hh[map62[term]].m_to_k[index][neighbour][2];
	    nn[4]=            hh[map62[term]].m_to_k[index][neighbour][3];
	  } else { /* 42 */
 	    /* k ---> m */
	    nn[1]=conf1[1][1]+hh[map62[term]].k_to_m[index][neighbour][0];
	    nn[2]=conf1[1][2]+hh[map62[term]].k_to_m[index][neighbour][1];
	    nn[3]=conf1[1][3]+hh[map62[term]].k_to_m[index][neighbour][2];
	    nn[4]=            hh[map62[term]].k_to_m[index][neighbour][3];
          }

	  numberFlips1=1;
	  for(m=1;m<=conf2_len;m++)
	    if (nn[1]==conf2[m][1] &&
		nn[2]==conf2[m][2] &&
		nn[3]==conf2[m][3] &&
		nn[4]==conf2[m][4])
	      numberFlips1++;
	
	  maxNumberFlips1=UnitCell.s2[nn[4]];
	
	  numberFlips2=0;
	  
	  for(m=1;m<=conf1_len;m++)
	    if (config0[k][1]==conf1[m][1] &&
		config0[k][2]==conf1[m][2] &&
		config0[k][3]==conf1[m][3] &&
		config0[k][4]==conf1[m][4])
	      numberFlips2++;
	
	  maxNumberFlips2=UnitCell.s2[config0[k][4]];

	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	
	  if (numberFlips1 <= maxNumberFlips1 &&
	      numberFlips2 <= maxNumberFlips2) {
	  	  
//	    temp_number_conf2_sites=conf2_len+1;
	  
	    conf2[conf2_len+1][1]=nn[1];
	    conf2[conf2_len+1][2]=nn[2];
	    conf2[conf2_len+1][3]=nn[3];
	    conf2[conf2_len+1][4]=nn[4];
	  
	    // H(m,m)
	    descriptorIndex1=findClusterIndex(conf1_len,conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    // F(k)
	    descriptorIndex2=findClusterIndex(conf2_len+1,conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    multiplier=1.0;
	    /*	    for(i=1;i<=(conf1_len-2);i++)
	      multiplier=multiplier*i; 
	      multiplier=multiplier*conf1_len*
	      (conf1_len-1);
	    */

	    // H(m,m)
	    // (length-2)!       -- because we have m in H(m,m) twice
	    //                      and we sum over it.
	    // length*(length-1) -- internal factor in H(m,m)

	    // for(i=1;i<conf2_len+1;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*(conf2_len+1);

	    // F(k)
	    // (length-1)!       -- permutation + we sum over k
	    // length            -- internal factor in F(k) 
	  
	    multiplier=(-2)*0.5*UnitCell.s2[nn[4]];

	    // factor from 2*S(z)(k)


	    {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		  hh[map62[term]].coefficient[index][neighbour];
                collect_term(indices,dummy,h_id);
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

// term41 = term39 backwards
// term42 = term40 backwards

double term43(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 43: H(k)(k) . H(m)(m) . S(+)(k) . S(+)(m)
   */
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  // double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  double sum;
  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips1=UnitCell.s2[config0[k1][4]];
	maxNumberFlips2=UnitCell.s2[config0[k2][4]];
	
	if (maxNumberFlips1 > 1 && maxNumberFlips2 > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=config0[k1][m]; // H(k,k) k--1st one
	    conf1[2][m]=config0[k1][m]; // H(k,k) k--2nd one
	    conf2[1][m]=config0[k2][m]; // H(m,m) m--1st one
	    conf2[2][m]=config0[k2][m]; // H(m,m) m--2nd one
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map62[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map62[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map62[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map62[term]].k_to_m[index][j][2];
	    nn[4]=hh[map62[term]].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    for(counter=0,l=1;l<=length;l++)
            if (l != k1 && l != k2 ) {
              counter++;
              tempArray[counter]=l;
	    }
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=2;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {      // H
		  conf1_len++;      temp=tempArray[j];
		  for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
		} else {            // H 
		  conf2_len++;      temp=tempArray[j];
		  for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (config0[k1][1]==conf1[m][1] &&
		    config0[k1][2]==conf1[m][2] &&
		    config0[k1][3]==conf1[m][3] &&
		    config0[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[config0[k1][4]];

	      numberFlips2=0;
	      
	      for(m=1;m<=conf2_len;m++) {
		if (config0[k2][1]==conf2[m][1] &&
		    config0[k2][2]==conf2[m][2] &&
		    config0[k2][3]==conf2[m][3] &&
		    config0[k2][4]==conf2[m][4])
		  numberFlips2++;
	      }
	      maxNumberFlips2=UnitCell.s2[config0[k2][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1 &&
		  numberFlips2 <= maxNumberFlips2) {
		
		descriptorIndex1=findClusterIndex(conf1_len,conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len,conf2);
                if (descriptorIndex2 <= 0) continue;

		// H(k,k)             
		// H(m,m)             

		{
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=
		      hh[map62[term]].coefficient[index][are_they_neighbours];
                    collect_term(indices,dummy,h_id);
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

// term45 is term44 backwards
double term44(int config_index, int length,
              int (*config0)[5], struct hamiltonian_defn hh[], int term)
{
  /*
     TERM44: H(k)(k) . F(m) . F(m) . S(+)(k) . S(+)(m)  2s>1
     TERM45: H(m)(m) . F(k) . F(k) . S(+)(k) . S(+)(m)  2s>1
  */
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int split0[MAX_LSUBM+1], i1;
  int k,l,m;
//  int j, mm,nnn,temp1,temp2,temp3, temp;
  int k1, k2;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  // double multiplier;
  int index;
//  int nn[5];
  int maxNumberFlips;
//  int numberFlips;
  int (*conf0)[5], (*conf1)[5], (*conf2)[5], (*conf3)[5];
  int conf1_len;
  int conf2_len;
  int conf3_len;
//  int tempArray[MAX_LSUBM+1];
  int counter;
//  int spin[MAX_LSUBM+2];
  int are_they_neighbours;

  double sum;
  sum=0.0000;

  conf0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
  
  for(k1=1;k1<=length;k1++)
  for(k2=1;k2<=length;k2++)
  if (k1 != k2) {
    
    index=config0[k1][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips <= 1) continue; // 2s>1
    
    for(m=1;m<5;m++) {
      conf1[1][m]=config0[k1][m]; // H(k,k) <-- 1st k here
      conf1[2][m]=config0[k1][m]; // H(k,k) <-- 2nd k here
      conf2[1][m]=config0[k2][m]; // F(m)
      conf3[1][m]=config0[k2][m]; // F(m)
    }
    
    // loop over nearest neighbours config0[k1], config0[k2]
    
    index=conf1[1][4];      
    are_they_neighbours=
      ((term==44)?test_k_to_m_neighbour(hh, term, conf1[1], conf2[1]):
       /* 45 */   test_m_to_k_neighbour(hh, term, conf1[1], conf2[1]));
    if (!are_they_neighbours) continue;
    
    for(counter=0,m=1;m<=length;m++)      // copy rest (len-2)
    if (m != k1 && m != k2) {
      counter++;
      conf0[counter][1]=config0[m][1]; // x
      conf0[counter][2]=config0[m][2]; // y
      conf0[counter][3]=config0[m][3]; // z
      conf0[counter][4]=config0[m][4]; // uc
    }
    
    // now we have num next neighbours in config possibilities
    // split the rest len-2 (3^(len-2) possibilities)
    //  but conf2 and conf3 are similar
    // we use trimary counter to split, zeros have to be maxlen-2
    //  num 1 can be more than num 2 if we set factor *= 2
      
    for (i1=1;i1<=counter;i1++) split0[i1]=0; // initiate splitter
    for (i1=0;!i1;) {

      conf1_len=2;
      conf2_len=1;
      conf3_len=1;

      i1=split_trimary(conf0,  counter, split0,
                       conf1,  &conf1_len,   // H(k,k)
                       conf2,  &conf2_len,   // F(m)
                       conf3,  &conf3_len);  // F(m)

#if 1 /* hit rarely!? no speed advantage! */
      if (to_much_flips(conf1, conf1_len, config0[k1]))
         continue;
#endif
  
      // H(k,k)
      descriptorIndex1=findClusterIndex(conf1_len,conf1);
      if (descriptorIndex1 <= 0) continue;
      
      // F(m)
      descriptorIndex2=findClusterIndex(conf2_len,conf2);
      if (descriptorIndex2 <= 0) continue;
      
      // F(m)
      descriptorIndex3=findClusterIndex(conf3_len,conf3);
      if (descriptorIndex3 <= 0) continue;
      
      // multiplier=1.0;
      {
        int indices[7]={0};
        double dummy;
        indices[0]=config_index;
        indices[1]=term;  /* 44 or 45 */
        indices[2]=descriptorIndex1;
        indices[3]=descriptorIndex2;
        indices[4]=descriptorIndex3;
        indices[5]=0;
        dummy=hh[map62[term]].coefficient[index][are_they_neighbours];
        collect_term(indices,dummy,h_id);
      }
    }
  }
  free(conf0);
  free(conf1);
  free(conf2);
  free(conf3);
  return sum;
}  


double term46(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
    TERM 46: - H(m)(m) . S(+)(m) . S(z)(k)
    TERM 50: - H(k)(k) . S(+)(k) . S(z)(m)
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptorIndex;
  double multiplier;
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips2;
  int numberFlips2;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  multiplier=1.0;

  for(k=1;k<=length;k++) {
    
    index=config0[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=config0[j][m]; // H(m,m) <- 1st m
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[length+1][m]=config0[k][m]; // H(m,m) <- 2nd m

      index=config0[k][4];
      
      for(neighbour=1;
          neighbour<=((term==46)?hh[map62[term]].m_to_k_number[index]:
                      /* 50 */   hh[map62[term]].k_to_m_number[index]);
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	if (term==46) {
  	  /* m ---> k */
	  nn[1]=config0[k][1]+hh[map62[term]].m_to_k[index][neighbour][0];
	  nn[2]=config0[k][2]+hh[map62[term]].m_to_k[index][neighbour][1];
	  nn[3]=config0[k][3]+hh[map62[term]].m_to_k[index][neighbour][2];
	  nn[4]=              hh[map62[term]].m_to_k[index][neighbour][3];
	} else { /* 50 */
  	  /* k ---> m */
	  nn[1]=config0[k][1]+hh[map62[term]].k_to_m[index][neighbour][0];
	  nn[2]=config0[k][2]+hh[map62[term]].k_to_m[index][neighbour][1];
	  nn[3]=config0[k][3]+hh[map62[term]].k_to_m[index][neighbour][2];
	  nn[4]=              hh[map62[term]].k_to_m[index][neighbour][3];
        }

	numberFlips2=0;
	for(l=1;l<=length+1;l++)
	  if (config0[k][1]==conf[l][1] &&
	      config0[k][2]==conf[l][2] &&
	      config0[k][3]==conf[l][3] &&
	      config0[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips2=UnitCell.s2[index];

	if (numberFlips2 <= maxNumberFlips2)  {
	  
	  descriptorIndex=findClusterIndex(length+1,conf);
	  
	  temp_multiplier=multiplier*0.5*UnitCell.s2[nn[4]];
	  
	  // coefficient from \tilde s^z \tilde s^- expansion
	  
          if (descriptorIndex > 0) {
            int indices[7]={0};
            double dummy;
            indices[0]=config_index;
            indices[1]=term;
            indices[2]=descriptorIndex;
            indices[3]=0;
            indices[4]=0;
            indices[5]=0;
            dummy=temp_multiplier*
              hh[map62[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
          }
	}
      }
    }
  }
  free(conf);
  return sum;
}

double term47(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 47: - M(k)(m)(m) . S(+)(k) . S(+)(m).
     TERM 49: - M(k)(k)(m) . S(+)(k) . S(+)(m).
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int i, j, k, l, m;
  int k1;
  int k2;
  int are_they_neighbours;
  int index, index1;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*conf)[5];
  double sum;
  sum=0.0000;

  conf= (int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
    
  // multiplier=1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k1 != k2) {

	for(i=1;i<=length;i++)
          for(j=1;j<5;j++) conf[i][j]=config0[i][j];
	
        for(m=1;m<5;m++) conf[length+1][m]=config0[k1][m]; // `m'
	
	numberFlips1=0;
	
	for(l=1;l<=length+1;l++) {
	  if (config0[k1][1]==conf[l][1] &&
	      config0[k1][2]==conf[l][2] &&
	      config0[k1][3]==conf[l][3] &&
	      config0[k1][4]==conf[l][4])
	    numberFlips1++;
	
	}

	index1=config0[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];

	index=config0[k1][4]; 
	are_they_neighbours=
          ((term==47)?test_m_to_k_neighbour(hh,term,config0[k1],config0[k2]):
	    /* 49 */  test_k_to_m_neighbour(hh,term,config0[k1],config0[k2]));

	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  descriptorIndex=findClusterIndex(length+1,conf);
	  
	  temp_multiplier=-1.0;

	  {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[map62[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	    }
	  }

	}
      }
    }
  }
  free(conf);
  return sum;
}

double term48(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
    TERM 48: - F(k) . H(m)(m) . S(+)(k) . S(+)(m)
    TERM 51: - F(m) . H(k)(k) . S(+)(k) . S(+)(m)
   */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int j, k, l, m;
  int k1;
  int k2;
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[config0[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=config0[k1][m]; // H(m,m) m--1st one
	    conf1[2][m]=config0[k1][m]; // H(m,m) m--2nd one
	    conf2[1][m]=config0[k2][m]; // F(k)   <- k
	  }
	  
	  // loop over nearest neighbours
	  
	  index=conf1[1][4];      
	  are_they_neighbours=
             ((term==48)?test_m_to_k_neighbour(hh,term,conf1[1],conf2[1]):
               /* 51 */  test_k_to_m_neighbour(hh,term,conf1[1],conf2[1]));
	  if (!are_they_neighbours) continue;

          {
	    
	    for(counter=0,l=1;l<=length;l++)
	    if (l != k1 && l != k2 ) {
		counter++;
		tempArray[counter]=l;
	    }
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {      // H
		  conf1_len++;      temp=tempArray[j];
		  for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
		} else {            // H
		  conf2_len++;      temp=tempArray[j];
		  for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (config0[k1][1]==conf1[m][1] &&
		    config0[k1][2]==conf1[m][2] &&
		    config0[k1][3]==conf1[m][3] &&
		    config0[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[config0[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		descriptorIndex1=findClusterIndex(conf1_len,conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len,conf2);
		if (descriptorIndex2 <= 0) continue;
		
		multiplier=1.0;
		/* for(i=1;i<=(conf1_len-2);i++) multiplier *= i;
		   multiplier *= conf1_len*(conf1_len-1);
		*/
		// H(m,m)             
		// (length-2)!        -- because we sum over m
		// length*(length-1)  -- factor due to M().
		/* for(i=1;i<conf2_len;i++) multiplier *= i;
		   multiplier *= conf2_len;
		*/
		// F(k)             
		// (length-1)!        -- because we sum over k
		// length             -- factor due to F().

		multiplier=-1.0; 

		// coefficient from \tilde s^- \tilde s^- expansion
		
		{
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map62[term]].coefficient[index][are_they_neighbours];
                    collect_term(indices,dummy,h_id);
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

// term49 = term47 backwards
// term50 = term46 backwards
// term51 = term48 backwards

double term52(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 52: - H(m)(m) . S(+)(k) . S(+)(m).
     TERM 53: - H(k)(k) . S(+)(k) . S(+)(m).
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int i, j, k, l, m;
  int k1;
  int k2;
  int counter;
  int are_they_neighbours;
  int index;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptorIndex;
  double multiplier;
  int (*conf)[5];
  double sum;
  sum=0.0000;

  conf=  (int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
       
  multiplier=-1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k1 != k2) {

	for(counter=1,i=1;i<=length;i++)
	if (i != k2) { 
	    for(j=1;j<5;j++) conf[counter][j]=config0[i][j];
	    counter++;
	}

	// now add in extra `m' site for H(m,m)
	
	for(m=1;m<5;m++) conf[length][m]=config0[k1][m]; // `m'
	
	numberFlips1=0;
	for(l=1;l<=length;l++) {
	  if (config0[k1][1]==conf[l][1] &&
	      config0[k1][2]==conf[l][2] &&
	      config0[k1][3]==conf[l][3] &&
	      config0[k1][4]==conf[l][4])
	    numberFlips1++;
	}

	index=config0[k1][4]; 
	maxNumberFlips1=UnitCell.s2[index];
	are_they_neighbours=
          ((term==52)?test_m_to_k_neighbour(hh,term,config0[k1],config0[k2]):
	   /* 53 */   test_k_to_m_neighbour(hh,term,config0[k1],config0[k2]));
	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {	  
	  
	  descriptorIndex=findClusterIndex(length,conf);
	  
	  {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[map62[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	    }
	  }

	}
      }
    }
  }
  free(conf);
  return sum;
}

// term53 = term52 backwards

// Single-site terms now: s^z

double term54(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
   /* 
     TERM 54: F(k) . S(+)(k)       Wed Jul 10 BST 2002
  */
  
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int k;
  int descriptorIndex;
  int index;
  double sum;
  sum=0.0000;

  descriptorIndex=findClusterIndex(length,config0);
  
  // JS 0903 this will cause error  (ToDo: should never happen?)
  //     for square_external_field.ccm + restriction!=4 
  //   x y uc [ 2]: 0 0 2, 0 0 1
  //  subsetClusters: stz=   -2 flag= 0
  //  Term 54 Error: descriptorIndex (0) != config_index (1)...
  //   -> restriction 3 (please check define.C!)
  if (descriptorIndex != config_index) {
    printf("Term 54 Error: descriptorIndex (%d"
      ") != config_index (%d)...\nExiting!!\n\n",
      descriptorIndex,config_index);
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  
  // JS0903: sum first, then call collect_term once (reduce input terms by n)

  for(k=1;k<=length;k++) {
    
    index=config0[k][4];                          // [4]=uc from ({},x,y,z,uc) 
    sum += hh[map62[term]].coefficient[index][1]; // single side term s^z
    
  }
  {
    int indices[7]={0};
    double dummy;
    indices[0]=config_index;
    indices[1]=term;
    indices[2]=config_index;
    indices[3]=0;
    indices[4]=0;
    indices[5]=0;
    dummy = sum;
    collect_term(indices,dummy,h_id);
  }
  
  return sum;
}

// Single-site: s^-

double term55(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 55: -2 * F(k) . S(z)(k) 
  */
  
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */
  // Oct05: configs with length+1 does not generate terms, so we skip it
  if (length+1>Params.n) return 0;

  int i, j, l;
  int descriptorIndex;
  double sum;
  sum=0.0000;
  
  int x,y,z,cell;
  int index;
  
  int numberFlips, maxNumberFlips;
  
  int (*config0)[5];
  
  config0=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      config0[i][j]=string[i][j];
  
  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
	  // order=length+1;   // this is the `k'    
	  
	  config0[length+1][1]=x;
	  config0[length+1][2]=y;
	  config0[length+1][3]=z;
	  config0[length+1][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  // check that we don't have too many spins at site `k'.
	  
	  for(l=1;l<=length;l++)
	    if (config0[length+1][1]==config0[l][1] &&
		config0[length+1][2]==config0[l][2] &&
		config0[length+1][3]==config0[l][3] &&
		config0[length+1][4]==config0[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[config0[length+1][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    // temp_multiplier=0.5*UnitCell.s2[config0[length+1][4]]*
	    //  multiplier;
	    
	    descriptorIndex=findClusterIndex(length+1,config0);
	    
            if (descriptorIndex > 0){
              int indices[7]={0};
              double dummy;
              indices[0]=config_index;
              indices[1]=term;
              indices[2]=descriptorIndex;
              indices[3]=0;
              indices[4]=0;
              indices[5]=0;
              dummy=(-2.0)*(-0.5*UnitCell.s2[config0[length+1][4]])*
                hh[map62[term]].coefficient[index][1];
              collect_term(indices,dummy,h_id);
            }
	  }
	}
      }
    }    
  }
  free(config0);
  return sum;  
}

double term56(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 56: - H(k)(k) . S(+)(k) 
  */
  
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptorIndex;
  // double multiplier;
  int (*conf)[5];
  int index;
  int maxNumberFlips;
  int maxNumberFlips2;
  int numberFlips2;
  double sum;
  sum=0.0000;
  
  conf= (int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
  
  // multiplier=1.0;
  
  for(k=1;k<=length;k++) {
    
    index=config0[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++) conf[j][m]=config0[j][m]; // H(m,m) <- 1st m
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++) conf[length+1][m]=config0[k][m]; // H(m,m) <- 2nd m
      
      index=config0[k][4];
      
      numberFlips2=0;
      
      for(l=1;l<=length+1;l++)
	if (config0[k][1]==conf[l][1] &&
	    config0[k][2]==conf[l][2] &&
	    config0[k][3]==conf[l][3] &&
	    config0[k][4]==conf[l][4])
	  numberFlips2++;
      
      maxNumberFlips2=UnitCell.s2[index];
      
      if (numberFlips2 <= maxNumberFlips2)  {
	
	descriptorIndex=findClusterIndex(length+1,conf);
	
	{
	  if (descriptorIndex > 0) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=term;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=(-1.0)*hh[map62[term]].coefficient[index][1];
            collect_term(indices,dummy,h_id);
	  }
	}
	
      }
    }
  }
  free(conf);
  return sum;
}

double term57(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
      TERM 57: - F(k) . F(k) . S(+)(k) 
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  int index;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  double sum;
  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) conf1[1][m]=config0[k][m]; // F(m)
    for(m=1;m<5;m++) conf2[1][m]=config0[k][m]; // F(m)

    for(counter=0,m=1;m<=length;m++)
    if (m != k) {
      counter++;
      tempArray[counter]=m;
    }
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {      // F(m) 1st one
	  conf1_len++;      temp=tempArray[j];
	  for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
	} else {            // F(m) 2nd one
	  conf2_len++;      temp=tempArray[j];
	  for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
	}
      }
      
      index=conf1[1][4];	

      descriptorIndex1=findClusterIndex(conf1_len,conf1);
      if (descriptorIndex1 <= 0) continue;
      
      descriptorIndex2=findClusterIndex(conf2_len,conf2);
      if (descriptorIndex2 <= 0) continue;

      {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=config_index;
	  indices[1]=term;
	  indices[2]=descriptorIndex1;
	  indices[3]=descriptorIndex2;
	  indices[4]=0;
	  indices[5]=0;
	  dummy=(-1.0)*hh[map62[term]].coefficient[index][1];
	  collect_term(indices,dummy,h_id);
      }
      
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

// s^z s^z : all s 

double term58(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /* 
     TERM 58: 2 F(k) . S(+)(k) . S(z)(k)
  */
  
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int k;
  int descriptorIndex;
  double multiplier;
  int index;
  double sum;
  sum=0.0000;
  
  // get the multiplying coefficient correct first!

  descriptorIndex=findClusterIndex(length,config0);
  
  if (descriptorIndex != config_index) {
    printf("Term 58 Error: descriptorIndex (%d"
      ") != config_index (%d)...\nExiting!!\n\n",
      descriptorIndex, config_index);
    exit(1);
  }
  
  multiplier=2.0;

  for(k=1;k<=length;k++) {
    
    index=config0[k][4];      
    
    // -0.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
    
    sum += -UnitCell.s2[index]*hh[map62[term]].coefficient[index][1];
  }
  // JS0903: moved outside loop, ToDo: very similar to 54?
  {
    int indices[7]={0};
    double dummy;
    indices[0]=config_index;
    indices[1]=term;
    indices[2]=config_index;
    indices[3]=0;
    indices[4]=0;
    indices[5]=0;
    dummy= sum;
    collect_term(indices,dummy,h_id);
  }
  return sum;
}

double term59(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 59: F(k) . S(+)(k) . S(+)(k).
  */


  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int i, j,  l;
  int k1;
  int k2;
  int are_they_the_same_lattice_site;
  int index;
  int descriptorIndex;
  double multiplier;
  int (*conf)[5];
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  conf= (int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k1 != k2) {
	
	for(l=1;l<5;l++) {    conf[1][l]=config0[k1][l]; } // F(m)
	for(l=1;l<5;l++) { tempConfig[l]=config0[k2][l]; } // s(k)^+
	
	are_they_the_same_lattice_site=0;
	index=conf[1][4];
	
	if(tempConfig[1] == conf[1][1] &&
	   tempConfig[2] == conf[1][2] &&
	   tempConfig[3] == conf[1][3] &&
	   tempConfig[4] == conf[1][4])
	  are_they_the_same_lattice_site=1;      
	  
	// now load the rest into conf
	
	for(counter=1,i=1;i<=length;i++)
	if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++) conf[counter][l]=config0[i][l];
	}
	
	if (counter != (length-1)) {
	  printf("TERM 59: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_the_same_lattice_site) {
	  
	  descriptorIndex=findClusterIndex(length-1,conf);
	  
          if (descriptorIndex > 0){
            int indices[7]={0};
            double dummy;
            indices[0]=config_index;
            indices[1]=term;
            indices[2]=descriptorIndex;
            indices[3]=0;
            indices[4]=0;
            indices[5]=0;
            dummy=multiplier*hh[map62[term]].coefficient[index][1];
            collect_term(indices,dummy,h_id);
          }
	}
      }
    }
  }
  free(conf);
  return sum;
}

double term60(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 60: H_{k,k} . S(+)(k) . S(+)(k).
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int i, j, l;
  int k1;
  int k2;
  int are_they_the_same_lattice_site;
  int index;
  int descriptorIndex;
  double multiplier;
  int (*conf)[5];
  int counter;
  double sum;
  sum=0.0000;

  conf= (int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k1 != k2) {
	
	for(l=1;l<5;l++) { conf[1][l]=config0[k1][l]; } // F(m)
	for(l=1;l<5;l++) { conf[2][l]=config0[k2][l]; } // s(k)^+
	
	are_they_the_same_lattice_site=0;
	index=conf[1][4];
	
	if(conf[2][1] == conf[1][1] &&
	   conf[2][2] == conf[1][2] &&
	   conf[2][3] == conf[1][3] &&
	   conf[2][4] == conf[1][4])
	  are_they_the_same_lattice_site=1;      
	  
	// now load the rest into conf
	
	for(counter=2,i=1;i<=length;i++)
	if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++) conf[counter][l]=config0[i][l];
	}
	
	if (!are_they_the_same_lattice_site) continue;

        {
	  
	  descriptorIndex=findClusterIndex(length,conf);
	  
	  {
	    if (descriptorIndex > 0){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[map62[term]].coefficient[index][1];
              collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf);
  return sum;
}

double term61(int config_index, int length, int (*config0)[5],
              struct hamiltonian_defn hh[], int term)
{
  /*
     TERM 61: F(k) . F(k) . S(+)(k) . S(+)(k).
  */

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int j, k, l, m;
  int k1;
  int k2;
  int are_they_the_same_lattice_site;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;

  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=config0[k1][m]; 
	for(m=1;m<5;m++) conf2[1][m]=config0[k2][m];
	
	// loop over nearest neighbours
	
	are_they_the_same_lattice_site=0;
	index=conf1[1][4];

	if(conf2[1][1] == conf1[1][1] &&
	   conf2[1][2] == conf1[1][2] &&
	   conf2[1][3] == conf1[1][3] &&
	   conf2[1][4] == conf1[1][4])
	  are_they_the_same_lattice_site=1;      
	
	if (!are_they_the_same_lattice_site) continue;

        {
	  
	  for(counter=0,l=1;l<=length;l++)
	  if (l != k1 && l != k2 ) {
	      counter++;
	      tempArray[counter]=l;
	  }	  
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) {
	      if (spin[j]) {
		conf1_len++;      temp=tempArray[j];
		for (m=1;m<5;m++) conf1[conf1_len][m]=config0[temp][m];
	      } else {
		conf2_len++;      temp=tempArray[j];
		for (m=1;m<5;m++) conf2[conf2_len][m]=config0[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findClusterIndex(conf1_len,conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findClusterIndex(conf2_len,conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    // now find the multiplying cofficient.
	    	    
	    {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=hh[map62[term]].coefficient[index][1];
		collect_term(indices,dummy,h_id);
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  return sum;
}

double term62(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 62: 
     
     S(+)(k)

     */
  if (!hh[map62[term]].evaluate) return 0; /* skip that term */
  
  int index;
  double sum;

  sum=0.0000;

  if (length == 1) {

    index=string[1][4];

    sum+=hh[map62[term]].coefficient[index][1];
    
    {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=term;
      indices[2]=0;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=hh[map62[term]].coefficient[index][1];
      collect_term(indices,dummy,h_id); 
    }
      
  }
  return sum;
}

/* Joerg Sep05:
 *  new: collect_gse_term() now collects gse data too
 * ToDo: estimate the size!
 */
// Joerg Jun08: obsolete
//void collect_gse_term(int indices[6],double dummy,int id){
//    collect_term(indices,dummy,id);
//}

/* now find terms which contribute to the ground-state energy */

double gse_term1(struct hamiltonian_defn hh[], int term)
{
 /* 
     GROUND-STATE ENERGY TERM 1:  term62=1 term13=1 numcoeff=0 
     
     GSE_TERM 1: S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int neighbour, index, new_index;
  double sum, multiplier;

  sum=0.0;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  // get the multiplying coefficient correct first!

  multiplier=0.25;

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[map62[term]].k_to_m_number[index];
	neighbour++) {
      
      new_index=hh[map62[term]].k_to_m[index][neighbour][3]; // uc (x,y,c,uc)
      
      sum += multiplier*UnitCell.s2[new_index]
                       *UnitCell.s2[index]*
	hh[map62[term]].coefficient[index][neighbour]/UnitCell.number;
      
    }
  }
  // JS0903: moved outside loop
  {
    int indices[7]={0};
    double dummy;
    indices[0]=0;
    indices[1]=term;
    indices[2]=0;
    indices[3]=0;
    indices[4]=0;
    indices[5]=0;
    dummy= sum;
    collect_term(indices,dummy,h_id);
  }
  return sum;
}
  
  
double gse_term2(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 2: term62=27 term13=9 
     
     4 G_{k,m} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, neighbour;
  int index, new_index;
  int descriptorIndex;
  double multiplier;
  int (*config0)[5];

  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  config0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      config0[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[map62[term]].k_to_m_number[index];
	neighbour++) {
      
      config0[1][1]=0;      // x
      config0[1][2]=0;      // y
      config0[1][3]=0;      // z
      config0[1][4]=index;  // uc
      
      config0[2][1]=config0[1][1]+hh[map62[term]].k_to_m[index][neighbour][0];
      config0[2][2]=config0[1][2]+hh[map62[term]].k_to_m[index][neighbour][1];
      config0[2][3]=config0[1][3]+hh[map62[term]].k_to_m[index][neighbour][2];
      config0[2][4]=              hh[map62[term]].k_to_m[index][neighbour][3];
      new_index=                  hh[map62[term]].k_to_m[index][neighbour][3];
      
      descriptorIndex=findClusterIndex(2,config0);
      
      multiplier=1.0;

      {
	if (descriptorIndex > 0) {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
          indices[1]=term;
	  indices[2]=descriptorIndex; // 2particle-nn-cluster
          indices[3]=0;
	  indices[4]=0;
          indices[5]=0;
	  dummy=multiplier*
	    UnitCell.s2[new_index]*
	    UnitCell.s2[index]*
	    hh[map62[term]].coefficient[index][neighbour]
	    /UnitCell.number;
          collect_term(indices,dummy,h_id);
	}
      }

    }
  }
  free(config0);
  return sum;
}
 
double gse_term3(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 3: term62=30 term13=9
     
    GS_TERM 3:  4 F_{k} F_{m} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, neighbour;
  int index, new_index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*config01)[5];
  int (*config02)[5];

  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  config01=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  config02=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) {
      config01[i][j]=0;
      config02[i][j]=0;
    }

  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[map62[term]].k_to_m_number[index];
	neighbour++) {
      
      config01[1][1]=0;
      config01[1][2]=0;
      config01[1][3]=0;
      config01[1][4]=index;
      
      new_index=hh[map62[term]].k_to_m[index][neighbour][3];
      
      config02[1][1]=config01[1][1]+hh[map62[term]].k_to_m[index][neighbour][0];
      config02[1][2]=config01[1][2]+hh[map62[term]].k_to_m[index][neighbour][1];
      config02[1][3]=config01[1][3]+hh[map62[term]].k_to_m[index][neighbour][2];
      config02[1][4]=                     hh[map62[term]].k_to_m[index][neighbour][3];
      
      descriptorIndex1=findClusterIndex(1,config01);
      if (descriptorIndex1 <= 0) continue;
      
      descriptorIndex2=findClusterIndex(1,config02);
      if (descriptorIndex2 <= 0) continue;
      
      multiplier = 4.0;
      multiplier *= (-0.5*UnitCell.s2[new_index])
                   *(-0.5*UnitCell.s2[index]);
      
      {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
	  indices[1]=term;
	  indices[2]=descriptorIndex1; // single site 0 - config
	  indices[3]=descriptorIndex2; // single neighbour site - config
	  indices[4]=0;
	  indices[5]=0;
	  dummy=multiplier*hh[map62[term]].coefficient[index][neighbour]
	    /UnitCell.number;
	  collect_term(indices,dummy,h_id);
      }
      
    }
  }
  free(config01);
  free(config02);
  return sum;
}
  
double gse_term4(struct hamiltonian_defn hh[], int term)
{
 /* 
     GROUND-STATE ENERGY TERM 4: term62=8 term13=3 
     
     GSE_TERM 4: -2 F_{m} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
  */

  int i, j, neighbour;
  int index, new_index;
  int descriptorIndex;
  double multiplier;
  int (*config0)[5];

  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  config0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      config0[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[map62[term]].k_to_m_number[index];
	neighbour++) {
      
      config0[1][1]=config0[1][1]+hh[map62[term]].k_to_m[index][neighbour][0];
      config0[1][2]=config0[1][2]+hh[map62[term]].k_to_m[index][neighbour][1];
      config0[1][3]=config0[1][3]+hh[map62[term]].k_to_m[index][neighbour][2];
      config0[1][4]=              hh[map62[term]].k_to_m[index][neighbour][3];
      new_index=                  hh[map62[term]].k_to_m[index][neighbour][3];
      
      descriptorIndex=findClusterIndex(1,config0);
      
      multiplier=-2.0;

      {
	if (descriptorIndex > 0) {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
          indices[1]=term;
	  indices[2]=descriptorIndex;
          indices[3]=0;
	  indices[4]=0;
          indices[5]=0;
	  dummy=multiplier*
	    (-0.5*UnitCell.s2[new_index])*
	    (-0.5*UnitCell.s2[index])*
	    hh[map62[term]].coefficient[index][neighbour]
	    /UnitCell.number;
	  collect_term(indices,dummy,h_id);
	}
      }

    }
  }
  free(config0);
  return sum;
}
  

double gse_term5(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 5: term62=17 term13=5 
     
     GSE_TERM 5: -2 F_{k} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, neighbour;
  int index, new_index;
  int descriptorIndex;
  double multiplier;
  int (*config0)[5];

  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  config0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      config0[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[map62[term]].k_to_m_number[index];
	neighbour++) {
      
      config0[1][1]=0;
      config0[1][2]=0;
      config0[1][3]=0;
      config0[1][4]=index;

      new_index=hh[map62[term]].k_to_m[index][neighbour][3];
      
      descriptorIndex=findClusterIndex(1,config0);
      
      multiplier=-2.0;

      {
	if (descriptorIndex > 0) {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
          indices[1]=term;
	  indices[2]=descriptorIndex;
          indices[3]=0;
	  indices[4]=0;
          indices[5]=0;
	  dummy=multiplier*
	    (-0.5*UnitCell.s2[new_index])*
	    (-0.5*UnitCell.s2[index])*
	    hh[map62[term]].coefficient[index][neighbour]
	    /UnitCell.number;
	  collect_term(indices,dummy,h_id);
	}
      }

    }
  }
  free(config0);
  return sum;
}
  
double gse_term6(struct hamiltonian_defn hh[], int term)
{
 /* 
     GROUND-STATE ENERGY TERM 6: Single Site, term62=54 term13=7
     
     GSE_TERM 6: S(z)(k) 

     Wed Jul 10 BST 2002
     
  */

  int index;
  double sum, multiplier;

  sum=0.0;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    multiplier=1.0;
    
    sum+=multiplier*(-0.5*UnitCell.s2[index])*
      hh[map62[term]].coefficient[index][1]/UnitCell.number;
    
    {
      int indices[7]={0};
      double dummy;
      indices[0]=0;
      indices[1]=term;
      indices[2]=0;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*
	(-0.5*UnitCell.s2[index])*
	hh[map62[term]].coefficient[index][1]
	/UnitCell.number;
      collect_term(indices,dummy,h_id);
    }
      
  }
  return sum;
}
  
double gse_term7(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 7: term62=55 term13=10 
     
     GSE_TERM 7: -2 F_{m} S(z)(k) 

     Wed Jul 10 BST 2002
     
  */

  int i, j;
  int index;
  int descriptorIndex;
  double multiplier;
  int (*config0)[5];

  double sum;
  sum=0.0000;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  config0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      config0[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    config0[1][1]=0;
    config0[1][2]=0;
    config0[1][3]=0;
    config0[1][4]=index;
      
    descriptorIndex=findClusterIndex(1,config0);
    
    multiplier=-2.0;
    
    {
      if (descriptorIndex > 0) {
	int indices[7]={0};
	double dummy;
	indices[0]=0;
        indices[1]=term;
	indices[2]=descriptorIndex;
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=multiplier*
	  (-0.5*UnitCell.s2[index])*
	  hh[map62[term]].coefficient[index][1]
	  /UnitCell.number;
	collect_term(indices,dummy,h_id);
      }
    }
    
  }
  free(config0);
  return sum;
}

double gse_term8(struct hamiltonian_defn hh[], int term)
{
 /* 
     GROUND-STATE ENERGY TERM 8: Single Site, term62=58 term13=11
     
     GSE_TERM 8: (S(z)(k))^2 

     Wed Jul 10 BST 2002
  */

  int index;
  double sum, multiplier;

  sum=0.0;

  if (!hh[map62[term]].evaluate) return 0; /* skip that term */

  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    multiplier=1.0;
    
    sum+=multiplier*(-0.5*UnitCell.s2[index])*(-0.5*UnitCell.s2[index])*
      hh[map62[term]].coefficient[index][1]/UnitCell.number;
    
    {
      int indices[7]={0};
      double dummy;
      indices[0]=0;
      indices[1]=term;
      indices[2]=0;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*
	(-0.5*UnitCell.s2[index])*
	(-0.5*UnitCell.s2[index])*
	hh[map62[term]].coefficient[index][1]
	/UnitCell.number;
      collect_term(indices,dummy,h_id);
    }
      
  }
  return sum;
}

int flag_subset=0; // Apr08 JS for SpeedUp

// same as obtainGenExp() ToDo: replace by one unique function!
void obtain(void)
{

  int    i,j,bonds;
  int    length;
  int    (*config0)[5];
  int    dummy;
  double energy;
  double lhs, rhs;
//  char output[60];
  time_t t1, t2;
  struct hamiltonian_defn *hh;  /* [MAX_TERMS13] */
  
//  static int n_out=0;   /* number of outputs */
#define MAX_out 10
  
//  char gs_name[100]="";

  config0=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  if (!config0) ERR("malloc config0");

  flag_subset=1;  // do subset-Check before findDescriptorIndex Apr08 JS

  if (!myid) printf("\n");

  // The data structures binary_max[], trimary_max[] ... are used 
  // to partition a given cluster into two (e.g. F(k) F(m)), three 
  // .... pieces in each `term' in H. 
  
  binary_max[0]=1;     dummy=2;   // 2^0 .. 2^30  # 30bit
  for(i=1;i<=30;i++) { binary_max[i]=dummy;   dummy=2*dummy; }

  trimary_max[0]=1;    dummy=3;   // 3^0 .. 3^10  # 16bit, i<=MAX_LSUBM
  for(i=1;i<=10;i++) { trimary_max[i]=dummy;  dummy=3*dummy; }

  // terms_fp=fopen(output,"w"); => done in ccm.C now

  // ground-state energy equation.

  // fp_gs=fopen(gs_name,"w");  /* same content for all threads */


  /* NEW!!: work out the maximum radius for all clusters 
     This is used to reduce the amount of sweeping for 
     F_k F_m s_k^z s_m^z -like terms. 
      v060611 JS: this function is moved here from input and define (copy)
  */
  radius=0;
  {
    int length;
    int (*tempConfig)[5]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
    for(i=1;i<=NumberConfigurations;i++) {
      length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
                                       tempConfig);
      /* note that midPointX, midPointY, midPointZ indicate Unit cells */
      // note that integer rounding shouldn't affect this 
      // algorithm
      // now find the maximum `radius'
      for(j=1;j<=length;j++) {
        if (abs(tempConfig[j][1]) > radius) radius=abs(tempConfig[j][1]);
        if (abs(tempConfig[j][2]) > radius) radius=abs(tempConfig[j][2]);
        if (abs(tempConfig[j][3]) > radius) radius=abs(tempConfig[j][3]);
      }
    }
    // radius+=1; bad if Dim<3 for(z=-radius,z<=radius,z++)  
    // ToDo: simpler maxdist m? or compute for each config (smaller)?
    if (!myid)
      printf("\nmaximum cell radius of all clusters = %d\n\n", radius+1);
    free(tempConfig);
  }

  // find CCM equations.

  if (!myid){
    printf(" Write: %s_gs %s.terms (buffered)\n\n",Control.name,Control.name);
    printf("==========================================================\n\n");
    printf("OBTAIN CCM EQUATIONS: START ALGORITHM.\n\n");
    printf("==========================================================\n\n");
  }

  lhs=0.0;rhs=0.0;energy=0.0;
  /* delta=1.0;  Joerg Sep05: delta seems to be not used here ??? */

  h_id=0; // used for collect_term, 0 means GSE-terms

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    hh=Hamiltonian[bonds];  // abbreviation
    
    // find ground-state energy, old:+=delta*gse_termX
    energy += gse_term1(hh, 1); // t13= 1 zz        Szk.Szm 0 
    energy += gse_term2(hh,27); // t13= 9 --   4Gkm.Szk.Szm 0,a1
    energy += gse_term3(hh,30); // t13= 9 -- 4Fk.Fm.Szk.Szm 0,a1,a2
    energy += gse_term4(hh, 8); // t13= 3 z-   -2Fm.Szk.Szm 0,a1
    energy += gse_term5(hh,17); // t13= 5 -z   -2Fk.Szk.Szm 0,a1
    energy += gse_term6(hh,54); // t13= 7 z             Szk 0 (single site)
    energy += gse_term7(hh,55); // t13=10 -        -2Fm.Szk 0,a1
    energy += gse_term8(hh,58); // t13=11 z^2         Szk^2 0
    {
      int indices[7]={0};
      indices[0]=STOP_CFG;indices[1]=STOP_TERM;  /* mark end of bond data */
      indices[2]=STOP_CFG;indices[3]=STOP_CFG;
      indices[4]=STOP_CFG;indices[5]=STOP_CFG;
      collect_term(indices,0.0,0);
    } 
    if (Control.output>0 && myid==0)
      printf("END GSE[bonds=%d] ===============================\n\n", bonds);
  }

  // fclose(fp_gs);

  /* This is very slow (only for bond2!) (Joerg 2005, ToDo) */
  t1=time(NULL);
  h_id=1; // used for collect_term

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    // n_out=0;  // reset for each bond    

    for(i=1;i<=NumberConfigurations;i++) { // total number of clusters

      /* not well balanced / scales badly (for 1D only, small #maxlencfgs)
       *       cpu time is not the same for each config -> balancing?
       * ToDo: one global counter for all threads, JS06
       *       let only one thread pass
       *       thread 0 starts with biggest config (=last cfg?)
       *       send counter-1 to next thread unblocked and start calculation
       *       receive counter and send -1 to next counter, stop if counter<0
       *       - send token to next ready process
       */
      if (((i-1)%npe)!=myid) continue;

      { // split loop to MPI threads
        
        static unsigned long nt_print=1; // used to output all 4^n

	length = findClusterFromDescriptor(spinConfigurations[i].descriptor,
		   			   config0);        
        if (Control.output > 2)
           printOutConfig(config0, length);

        { // do it once! used in term8,17,27[1],30[2],55
           int maxPointX, maxPointY, maxPointZ;

           // ToDo: do outside outer loop
           maxPointX=maxPointY=maxPointZ=0;
           for(j=1;j<=length;j++) {
             if (maxPointX < abs(config0[j][1])) maxPointX=abs(config0[j][1]);
             if (maxPointY < abs(config0[j][2])) maxPointY=abs(config0[j][2]);
             if (maxPointZ < abs(config0[j][3])) maxPointZ=abs(config0[j][3]);
           }
           // old
           //  minX=radius;           minY=radius;           minZ=radius;
           //  maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=radius+maxPointZ;
           // new (0903 wrong for square_external_field.ccm)
           // was bad, why?
           // this is optimal for term 8, 17, 27 and 55 [1] (no splitting)
           minX=radius-maxPointX+1; minY=radius-maxPointY+1; minZ=radius-maxPointZ+1;
           maxX=radius+1;           maxY=radius+1;           maxZ=radius+1;
           // but term30 [2] does splitting and xyz has bigger range
           // smaller range gives wrong results for
           //    square_external_field.ccm
           //  good iterms30=2277 oterms09=480 -0.663637108485
           //  bad  iterms30=2206 iterms09=474 -0.663612145867 v0902
           //
           minX=radius+1;           minY=radius+1;           minZ=radius+1;
           maxX=radius+maxPointX+1; maxY=radius+maxPointY+1; maxZ=radius+maxPointZ+1;
 
           // range for additional spin flip
           if (Params.dim == 1) {
             minY=0;                  minZ=0;
             maxY=0;                  maxZ=0;
           } else if (Params.dim == 2) {
             minZ=0;
             maxZ=0;
           }
           if (Params.dim > 3) {
             printf("Really weird error. Dimension (%d)"
               " not recognised...\nExiting||\n\n", Params.dim);
             exit(1);
           }
           // JS: better minX=radius-maxPointX, maxX=radius;
           //    dim=2: minZ=0 maxZ=0
           // JS: makes no sence? because all configs have maxX = "radius"
        }
        // we want to output not to much because not every user knows
        //    what he does ;)
        // ToDo: output only for numterms=1,4,16,64,256,1024,4096,...
	if (Control.output > 0 && myid == 0
         && nterms_out1 > nt_print) {  // we only output every 4^n
          t2=time(NULL)-t1;
          if (nt_print<<1) nt_print<<=1;  // *4 but avoid overflow
          if (t2>27)  // Apr08 reduce output
          // nterms_out1 is number terms of this thread 
          printf("obtain: Bond %d : Cluster<= %8d %6.2f%%"
                 " : terms/cluster= %8.2f %5.1fm\n",
                 bonds, i, i*100./NumberConfigurations,
                 npe*1.*nterms_out1/(i+1.), t2/60.);
          fflush(stdout); // no buffering (possible time estimations)
        }

        hh=Hamiltonian[bonds];  // abbreviation

        // there are 13 terms defined in .ccm, expanded here to 62 terms
        //
	
	// s^z s^z : s all s = t1  // square, lc1, lc_mf (x,y,z,uc,r)
	
        // term1 ???
	lhs += term2(i,length,config0,hh,2);  // map62[1..5]=1
	lhs += term2(i,length,config0,hh,3);
	lhs += term4(i,length,config0,hh,4);
	lhs += term5(i,length,config0,hh,5);
	
	// s^z s^+ all s = t2 (2*)
	
	rhs += term6(i,length,config0,hh,6);  // map62[6..7]=2
	rhs += term7(i,length,config0,hh,7);
      
	// s^z s^- all s = t3 (6*)   // square+field, quasi2Dj1j2j3, lc_mf
	
	rhs += term8(i,length,config0,hh,8); // map62[8..13]=3
	rhs += term9(i,length,config0,hh,9);
	rhs += term10(i,length,config0,hh,10);
	rhs += term11(i,length,config0,hh,11);
	rhs += term12(i,length,config0,hh,12);
	rhs += term13(i,length,config0,hh,13);
	
        // s^z s^-  2s>1  = t3 (collect better t3+t3, JS0902)
	rhs += term46(i,length,config0,hh,46); // map62[46]=3
	rhs += term47(i,length,config0,hh,47);
	rhs += term48(i,length,config0,hh,48);

	// s^+ s^z all s = t4 (2*)
	
	rhs += term6(i,length,config0,hh,14); // map62[14..15]=4
	rhs += term7(i,length,config0,hh,15);
      
        // s^- s^z all s = t5 (6*)   // square+field, quasi2Dj1j2j3, lc_mf
        //  replaced by t3 variant!
	rhs += term9(i,length,config0,hh,16); 
	rhs += term8(i,length,config0,hh,17);
	rhs += term12(i,length,config0,hh,18);
	rhs += term11(i,length,config0,hh,19);
	rhs += term10(i,length,config0,hh,20);
	rhs += term13(i,length,config0,hh,21); // JS0902

        // s^- s^z 2s>1  = t5
	rhs += term47(i,length,config0,hh,49); // map62[49]=5
	rhs += term46(i,length,config0,hh,50);
	rhs += term48(i,length,config0,hh,51);

        // s^+ s^- all s = t6 (2*)
	rhs += term22(i,length,config0,hh,22);
	rhs += term23(i,length,config0,hh,23);

        // s^+ s^- 2s>1 = t6
	rhs += term52(i,length,config0,hh,52); // map62[52]=6

        // s^- s^+ all s = t7 (2*)
	rhs += term22(i,length,config0,hh,24);
	rhs += term23(i,length,config0,hh,25);

        // s^- s^+ 2s>1 = t7
	rhs += term52(i,length,config0,hh,53); // map62[53]=7

        // s^+ s^+ all s = t8 (1*) // lc2 (x,y,z,uc,r)
	rhs += term26(i,length,config0,hh,26); // map62[26]=8
      
        // s^- s^- all s = t9 (9*) // lc2 (x,y,z,uc,r)
	rhs += term27(i,length,config0,hh,27); /* => length+1 */
	rhs += term28(i,length,config0,hh,28);
	rhs += term29(i,length,config0,hh,29);
	rhs += term30(i,length,config0,hh,30); /* slowest 75% square */
	rhs += term31(i,length,config0,hh,31); /* t30..34 90% of slowness */
	rhs += term32(i,length,config0,hh,32);
	rhs += term31(i,length,config0,hh,33); // JS0902 replaced
	rhs += term34(i,length,config0,hh,34); /* slowest 31% chain */
	rhs += term29(i,length,config0,hh,35);

        // s^- s^-  2s>1  = t9 (10*) // lc2 (x,y,z,uc,r)
	rhs += term36(i,length,config0,hh,36); // map62[36]=9
	rhs += term36(i,length,config0,hh,37); // JS0902 replaced
	rhs += term38(i,length,config0,hh,38);
	rhs += term39(i,length,config0,hh,39);
	rhs += term40(i,length,config0,hh,40);
	rhs += term39(i,length,config0,hh,41);
	rhs += term40(i,length,config0,hh,42);
	rhs += term43(i,length,config0,hh,43);
	rhs += term44(i,length,config0,hh,44);
	rhs += term44(i,length,config0,hh,45); // JS0902 replaced
      
        // s^z all s = t10  // lc12 (r=0)
	rhs += term54(i,length,config0,hh,54); // map62=10
     
        // s^- all s = t11  // lc12 (r=0)
	rhs += term55(i,length,config0,hh,55); // map62=11
	rhs += term56(i,length,config0,hh,56);
	rhs += term57(i,length,config0,hh,57);

        // (s^z)^2 = t12    // lc12 (r=0)
	rhs += term58(i,length,config0,hh,58); // map62=12
	rhs += term59(i,length,config0,hh,59);
	rhs += term60(i,length,config0,hh,60);
	rhs += term61(i,length,config0,hh,61);

        // s^+     = t13    // lc12 (r=0)
	rhs += term62(i,length,config0,hh,62); // map62=13
      }
    }
    {
      int indices[7]={0};
      indices[0]=STOP_CFG; indices[1]=STOP_TERM;
      indices[2]=STOP_CFG; indices[3]=STOP_CFG;
      indices[4]=STOP_CFG; indices[5]=STOP_CFG;
      collect_term(indices,0.0,1);
    }
    if (Control.output>-1 && myid==0)
      printf("END H[bonds=%d] ==================================\n\n",bonds);
  }
  
  free(config0); // memleak until v0606

  t2=time(NULL)-t1; // ToDo: min/max(MPI)
  // fclose(terms_fp); => closed in ccm.C now

  if (myid==0) {
    printf("\nOBTAIN CCM EQUATIONS: SUCCESSFULLY COMPLETED!\n\n");
    printf("obtain.t[min]=%.2f (jobid=0)\n", t2/60.);
  }

  return;
}
